from uuid import UUID, uuid4

import django.apps
import pytest
from django.core.management import call_command
from django.db import connection
from django.db.migrations.executor import MigrationExecutor
from django.db.utils import DataError
from django.utils import timezone

from caluma.utils import col_type_from_db


def test_migrate_to_flat_answers(transactional_db):
    executor = MigrationExecutor(connection)
    app = "caluma_form"
    migrate_from = [(app, "0017_auto_20190619_1320")]
    migrate_to = [(app, "0019_remove_answer_value_document")]

    executor.migrate(migrate_from)
    old_apps = executor.loader.project_state(migrate_from).apps

    # Create some old data. Can't use factories here

    Document = old_apps.get_model(app, "Document")
    Form = old_apps.get_model(app, "Form")
    Answer = old_apps.get_model(app, "Answer")
    Question = old_apps.get_model(app, "Question")
    FormQuestion = old_apps.get_model(app, "FormQuestion")

    main_form = Form.objects.create(slug="main-form")
    sub_form = Form.objects.create(slug="sub-form")

    main_form_question = Question.objects.create(
        type="form", sub_form=sub_form, slug="main-form-question"
    )
    FormQuestion.objects.create(form=main_form, question=main_form_question)

    main_text_question = Question.objects.create(type="text", slug="main-text-question")
    FormQuestion.objects.create(form=main_form, question=main_text_question)

    sub_text_question = Question.objects.create(type="text", slug="sub_1_question_1")
    FormQuestion.objects.create(form=sub_form, question=sub_text_question)

    # Note: in this step, Document.family is (historically) a UUID4 field, and
    # not yet a foreign key.
    # We need to set a temporary family, because the signals are not available...
    main_document = Document.objects.create(form=main_form, family=uuid4())
    # ... then we set the correct family
    main_document.family = main_document.pk
    main_document.save()

    sub_document = Document.objects.create(form=sub_form, family=main_document.pk)
    sub_answer = Answer.objects.create(
        value="lorem ipsum", question=sub_text_question, document=sub_document
    )
    Answer.objects.create(
        value_document=sub_document, question=main_form_question, document=main_document
    )

    text_answer = Answer.objects.create(
        value="dolor sit", question=main_text_question, document=main_document
    )

    assert not sub_answer.document == main_document
    assert text_answer.document == main_document

    # Migrate forwards.
    executor.loader.build_graph()  # reload.
    executor.migrate(migrate_to)
    new_apps = executor.loader.project_state(migrate_to).apps

    # Test the new data.
    Answer = new_apps.get_model(app, "Answer")

    sub_answer = Answer.objects.get(value="lorem ipsum")
    text_answer = Answer.objects.get(value="dolor sit")

    assert Answer.objects.filter(question__type="form").count() == 0
    assert sub_answer.document.pk == main_document.pk
    assert text_answer.document.pk == main_document.pk


def test_migrate_to_form_question_natural_key_forward(transactional_db):
    executor = MigrationExecutor(connection)
    app = "caluma_form"
    migrate_from = [(app, "0023_auto_20190729_1448")]
    migrate_to = [(app, "0024_auto_20190919_1244")]

    executor.migrate(migrate_from)
    old_apps = executor.loader.project_state(migrate_from).apps

    # Create some old data. Can't use factories here

    Form = old_apps.get_model(app, "Form")
    Question = old_apps.get_model(app, "Question")
    FormQuestion = old_apps.get_model(app, "FormQuestion")

    form_1 = Form.objects.create(slug="form-1")

    question_1 = Question.objects.create(type="text", slug="question-1")
    form_question = FormQuestion.objects.create(form=form_1, question=question_1)

    assert isinstance(form_question.pk, UUID)

    # Migrate forwards.
    executor.loader.build_graph()  # reload.
    executor.migrate(migrate_to)
    new_apps = executor.loader.project_state(migrate_to).apps

    # Test the new data.
    FormQuestion = new_apps.get_model(app, "FormQuestion")

    form_question = FormQuestion.objects.first()

    assert form_question.pk == "form-1.question-1"


def test_migrate_to_form_question_natural_key_reverse(transactional_db):
    executor = MigrationExecutor(connection)
    app = "caluma_form"
    migrate_from = [(app, "0024_auto_20190919_1244")]
    migrate_to = [(app, "0023_auto_20190729_1448")]

    executor.migrate(migrate_from)
    old_apps = executor.loader.project_state(migrate_from).apps

    # Create some old data. Can't use factories here

    Form = old_apps.get_model(app, "Form")
    Question = old_apps.get_model(app, "Question")
    FormQuestion = old_apps.get_model(app, "FormQuestion")

    form_1 = Form.objects.create(slug="form-1")

    question_1 = Question.objects.create(type="text", slug="question-1")
    FormQuestion.objects.create(form=form_1, question=question_1)

    # Migrate backwards.
    executor.loader.build_graph()  # reload.
    with pytest.raises(DataError):
        executor.migrate(migrate_to)


def test_dynamic_option_unique_together(transactional_db):
    executor = MigrationExecutor(connection)
    app = "caluma_form"
    migrate_from = [(app, "0028_auto_20200210_0929")]
    migrate_to = [(app, "0029_dynamic_option_unique_together")]

    executor.migrate(migrate_from)
    old_apps = executor.loader.project_state(migrate_from).apps

    # Create some old data. Can't use factories here

    Document = old_apps.get_model(app, "Document")
    Form = old_apps.get_model(app, "Form")
    Question = old_apps.get_model(app, "Question")
    DynamicOption = old_apps.get_model(app, "DynamicOption")
    HistoricalDynamicOption = old_apps.get_model(
        "caluma_form", "HistoricalDynamicOption"
    )

    form = Form.objects.create(slug="main-form")

    question = Question.objects.create(type="text", slug="main-form-question")

    # we need to set a temporary family, because the signals are not available
    document = Document.objects.create(form=form, family=uuid4())
    # then we set the correct family
    document.family = document.pk
    document.save()

    d_option_1 = DynamicOption.objects.create(
        document=document, question=question, slug="foo"
    )
    d_option_2 = DynamicOption.objects.create(
        document=document, question=question, slug="foo"
    )

    # again no signals
    HistoricalDynamicOption.objects.create(
        document=document,
        question=question,
        history_type="+",
        created_at=d_option_1.created_at,
        modified_at=d_option_1.modified_at,
        history_date=d_option_1.modified_at,
        slug="foo",
        id=d_option_1.pk,
    )
    HistoricalDynamicOption.objects.create(
        document=document,
        question=question,
        history_type="+",
        created_at=d_option_2.created_at,
        modified_at=d_option_2.modified_at,
        history_date=d_option_2.modified_at,
        slug="foo",
        id=d_option_2.pk,
    )

    # Migrate forwards.
    executor.loader.build_graph()  # reload.
    executor.migrate(migrate_to)
    new_apps = executor.loader.project_state(migrate_to).apps

    # Test the new data.
    DynamicOption = new_apps.get_model(app, "DynamicOption")
    HistoricalDynamicOption = new_apps.get_model(
        "caluma_form", "HistoricalDynamicOption"
    )
    Document = new_apps.get_model(app, "Document")
    Question = new_apps.get_model(app, "Question")

    document = Document.objects.get(pk=document.pk)
    question = Question.objects.get(pk=question.pk)

    assert DynamicOption.objects.get(document=document, question=question, slug="foo")
    assert (
        HistoricalDynamicOption.objects.filter(
            document=document, question=question, slug="foo"
        ).count()
        == 1
    )


def test_migrate_to_family_as_pk(transactional_db, caplog):
    """Ensure correct behaviour when moving family to PK.

    Document family may not match an existing document.
    In this case, the family should be set to their own PK.
    """
    executor = MigrationExecutor(connection)
    app = "caluma_form"
    migrate_from = [(app, "0030_auto_20200219_1359")]
    migrate_to = [(app, "0031_auto_20200220_0910")]

    executor.migrate(migrate_from)
    old_apps = executor.loader.project_state(migrate_from).apps

    OldDocument = old_apps.get_model(app, "Document")
    OldForm = old_apps.get_model(app, "Form")

    form = OldForm.objects.create(slug="form-1")

    # We don't have any signals here, so have to do it ourselves
    root_doc = OldDocument.objects.create(form=form, family=uuid4())
    root_doc.family = root_doc.pk
    root_doc.save()

    other_doc = OldDocument.objects.create(form=form, family=root_doc.pk)
    unrelated_doc = OldDocument.objects.create(form=form, family=uuid4())

    # migration should log that our unrelated_doc needs fixing
    expected_msg = (
        f"Document pk={unrelated_doc.pk} (form={form.pk}) "
        f"missing it's family={unrelated_doc.family}. "
        f"Resetting to itself"
    )

    assert isinstance(other_doc.family, UUID)

    # Migrate forwards.
    executor.loader.build_graph()  # reload.
    executor.migrate(migrate_to)

    new_apps = executor.loader.project_state(migrate_to).apps

    # Test the new data.
    NewDocument = new_apps.get_model(app, "Document")

    new_root_doc = NewDocument.objects.get(pk=root_doc.pk)
    new_other_doc = NewDocument.objects.get(pk=other_doc.pk)
    new_unrelated_doc = NewDocument.objects.get(pk=unrelated_doc.pk)

    assert new_other_doc.family == new_root_doc
    assert new_unrelated_doc.family == new_unrelated_doc
    assert new_root_doc.family == new_root_doc
    assert expected_msg in caplog.messages


def _verify_foreign_key_types(apps):

    for models in apps.all_models.values():
        for model in models.values():
            slug_fks = [
                field
                for field in model._meta.fields
                if field.is_relation and field.target_field.name == "slug"
            ]

            for field in slug_fks:
                # ok we have a slug foreign key
                fk_params = field.db_parameters(connection)
                target_params = field.target_field.db_parameters(connection)

                # verify django-internal specified type
                assert (
                    fk_params["type"] == target_params["type"]
                ), f"Foreign key field {field}: type mismatch with destination in django-internal representation"

                # check if the DB agrees
                fk_dbtype = col_type_from_db(field, connection)
                target_dbtype = col_type_from_db(field.target_field, connection)
                assert (
                    fk_dbtype == target_dbtype
                ), f"Foreign key field {field}: type mismatch with destination in DB"


def test_slugfield_length_correctness(transactional_db):
    """Detect deviation of foreign key types from target field types.

    Note: If this test fails, you'll most likely need to create a new
    migration and run caluma.utils.fix_foreign_key_types() in it to
    cleanup the mess again.
    """

    # Just make sure we're at the newest version, as we're messing
    # with migrations in these tests around here
    call_command("migrate", no_input=True)
    _verify_foreign_key_types(django.apps.apps)


def test_migrate_slugfield_length(transactional_db):
    """Ensure migration of slugfield length works correctly."""

    # we need to migrate down and up again to consistently trigger
    # the type mismatch :(

    migration_back = [
        # first change (50 -> 150)
        ("caluma_form", "0032_auto_20200220_1311"),
        ("caluma_workflow", "0018_auto_20200219_1359"),
    ]

    migration_with_fixes = [("caluma_form", "0034_fix_fk_lengths")]

    executor = MigrationExecutor(connection)

    apps = executor.migrate(migration_back).apps

    with pytest.raises(AssertionError):
        _verify_foreign_key_types(apps)

    executor.loader.build_graph()  # reload.
    new_apps = executor.migrate(migration_with_fixes).apps

    # should now throw anymore
    _verify_foreign_key_types(new_apps)


def test_migrate_answer_history_question_type(transactional_db):
    """Make sure migration to custom history field history_question_type works."""

    executor = MigrationExecutor(connection)
    app = "caluma_form"
    migrate_from = [(app, "0034_fix_fk_lengths")]
    migrate_to = [(app, "0035_historicalanswer_history_question_type")]

    executor.migrate(migrate_from)
    old_apps = executor.loader.project_state(migrate_from).apps

    Document = old_apps.get_model(app, "Document")
    Form = old_apps.get_model(app, "Form")
    OldAnswer = old_apps.get_model(app, "Answer")
    OldQuestion = old_apps.get_model(app, "Question")
    OldHistAns = old_apps.get_model(app, "HistoricalAnswer")
    OldHistQuest = old_apps.get_model(app, "HistoricalQuestion")

    form = Form.objects.create()
    question = OldQuestion.objects.create(type="text")
    document = Document.objects.create(form=form)
    answer = OldAnswer.objects.create(
        document=document, question=question, value="some answer"
    )

    now = timezone.now()

    # create historical records
    old_hist_quest = OldHistQuest.objects.create(
        slug=question.slug,
        type="text",
        created_at=now,
        modified_at=now,
        history_date=now,
    )
    old_hist_ans = OldHistAns.objects.create(
        id=answer.id,
        value=answer.value,
        document_id=document.pk,
        question_id=question.slug,
        created_at=now,
        modified_at=now,
        history_date=now,
    )
    # create another set of question / answer
    now = timezone.now()

    new_hist_quest = OldHistQuest.objects.create(
        slug=question.slug,
        type="integer",
        created_at=now,
        modified_at=now,
        history_date=now,
    )
    new_hist_ans = OldHistAns.objects.create(
        id=answer.id,
        value=answer.value,
        document_id=document.pk,
        question_id=question.slug,
        created_at=now,
        modified_at=now,
        history_date=now,
    )
    # Migrate forward
    executor.loader.build_graph()
    executor.migrate(migrate_to)

    new_apps = executor.loader.project_state(migrate_to).apps

    MigratedHistAns = new_apps.get_model(app, "HistoricalAnswer")

    # Test the new data
    old_hist_ans = MigratedHistAns.objects.get(history_id=old_hist_ans.history_id)
    assert old_hist_ans.history_question_type == old_hist_quest.type

    new_hist_ans = MigratedHistAns.objects.get(history_id=new_hist_ans.history_id)
    assert new_hist_ans.history_question_type == new_hist_quest.type



import pytest

from ...caluma_core.models import UUIDModel
from ...caluma_core.tests.fake_model import get_fake_model
from .. import permissions


@pytest.mark.parametrize(
    "info_fixture,is_authenticated", [("info", False), ("admin_info", True)]
)
def test_is_authenticated_permission(db, info_fixture, is_authenticated, request):
    info = request.getfixturevalue(info_fixture)
    assert permissions.IsAuthenticated().has_permission(None, info) == is_authenticated


@pytest.mark.parametrize(
    "admin_groups,is_created_by", [(["admin_group"], True), (["nogroup"], False)]
)
def test_created_by_group_permission(db, admin_info, is_created_by, history_mock):
    FakeModel = get_fake_model(model_base=UUIDModel)
    instance = FakeModel.objects.create(created_by_group="admin_group")
    assert (
        permissions.CreatedByGroup().has_object_permission(None, admin_info, instance)
        == is_created_by
    )



from django.contrib.postgres.fields import JSONField
from django.db import models


class AccessLog(models.Model):
    OPERATIONS = (("query", "query"), ("mutation", "mutation"))

    timestamp = models.DateTimeField(auto_now_add=True)
    username = models.CharField(max_length=100, blank=True, null=True)
    query = models.TextField()
    operation = models.CharField(choices=OPERATIONS, max_length=20)
    operation_name = models.CharField(max_length=100, blank=True, null=True)
    selection = models.CharField(max_length=100, blank=True, null=True)
    variables = JSONField(blank=True, null=True)
    status_code = models.PositiveIntegerField()
    has_error = models.BooleanField(default=False)



import pytest
from django.core.management import call_command
from django.core.management.base import CommandError

from caluma.caluma_form.models import Form


@pytest.mark.parametrize(
    "form__slug,form__name,form__description,form__is_published",
    [("form-v1", "Form v1", "First version of the form", True)],
)
@pytest.mark.parametrize(
    "params,expected_form",
    [
        (
            {"slug": "form-v2"},
            {
                "slug": "form-v2",
                "name": "",
                "description": "",
                "is_published": False,
            },
        ),
        (
            {
                "slug": "form-v2",
                "name": "Form v2",
                "description": "Second version of the form",
                "is_published": True,
            },
            {
                "slug": "form-v2",
                "name": "Form v2",
                "description": "Second version of the form",
                "is_published": True,
            },
        ),
    ],
)
def test_copy_form_command(db, form, params, expected_form):
    call_command("copy_form", form.pk, **params)

    new_form = Form.objects.get(pk=params["slug"])

    for key, value in expected_form.items():
        assert getattr(new_form, key) == value


def test_copy_form_command_error(db):
    with pytest.raises(CommandError) as exc:
        call_command("copy_form", "nonexistent", slug="new-slug")

    assert str(exc.value) == "Form 'nonexistent' not found - can't copy"



from typing import Optional

from django.db import transaction
from django.db.models import Q
from rest_framework.exceptions import ValidationError

from caluma.caluma_core.models import BaseModel
from caluma.caluma_form import models, validators
from caluma.caluma_user.models import BaseUser
from caluma.utils import update_model


class BaseLogic:
    @staticmethod
    @transaction.atomic
    def create(model: BaseModel, validated_data, user: Optional[BaseUser] = None):
        if user:
            validated_data["created_by_user"] = user.username if user else None
            validated_data["created_by_group"] = user.group if user else None
            validated_data["modified_by_user"] = user.username if user else None
            validated_data["modified_by_group"] = user.group if user else None

        return model.objects.create(**validated_data)

    @staticmethod
    @transaction.atomic
    def update(instance: BaseModel, validated_data, user: Optional[BaseUser] = None):
        if user:
            validated_data["modified_by_user"] = user.username
            validated_data["modified_by_group"] = user.group

        update_model(instance, validated_data)
        instance.refresh_from_db()

        return instance


class SaveAnswerLogic:
    @classmethod
    def get_new_answer(cls, data, user, answer):
        validated_data = cls.pre_save(
            cls.validate_for_save(data, user, answer=answer, origin=True)
        )

        if answer is None:
            answer = cls.create(validated_data, user)
        else:
            answer = cls.update(answer, validated_data, user)

        return cls.post_save(answer)

    @staticmethod
    def validate_for_save(
        data: dict, user: BaseUser, answer: models.Answer = None, origin: bool = False
    ) -> dict:
        question = data["question"]

        # Save empty answer by omitting the value property since graphql-core
        # doesn't allow `null` literals yet
        if "value" not in data:
            data["value"] = None

        # make data from python api similar to graphql api
        if question.type == models.Question.TYPE_TABLE and not data.get("documents"):
            data["documents"] = (
                models.Document.objects.filter(pk__in=data["value"])
                if data["value"]
                else models.Document.objects.none()
            )
            del data["value"]
        elif question.type == models.Question.TYPE_FILE and not data.get("file"):
            data["file"] = data["value"]
            del data["value"]
        elif question.type == models.Question.TYPE_DATE and not data.get("date"):
            data["date"] = data["value"]
            del data["value"]

        validators.AnswerValidator().validate(
            **data, user=user, instance=answer, origin=origin
        )

        return data

    @staticmethod
    def pre_save(validated_data: dict) -> dict:
        # TODO emit events
        return validated_data

    @staticmethod
    def post_save(answer: models.Answer) -> models.Answer:
        # TODO emit events
        return answer

    @staticmethod
    @transaction.atomic
    def create(validated_data: dict, user: Optional[BaseUser] = None) -> models.Answer:
        if validated_data["question"].type == models.Question.TYPE_FILE:
            validated_data = __class__.set_file(validated_data)

        if validated_data["question"].type == models.Question.TYPE_TABLE:
            documents = validated_data.pop("documents")

        answer = BaseLogic.create(models.Answer, validated_data, user)

        if answer.question.type == models.Question.TYPE_TABLE:
            answer.create_answer_documents(documents)

        return answer

    @staticmethod
    @transaction.atomic
    def update(answer, validated_data, user: Optional[BaseUser] = None):
        if answer.question.type == models.Question.TYPE_TABLE:
            documents = validated_data.pop("documents")
            answer.unlink_unused_rows(docs_to_keep=documents)

        if (
            answer.question.type == models.Question.TYPE_FILE
            and answer.file.name is not validated_data["file"]
        ):
            answer.file.delete()
            validated_data = __class__.set_file(validated_data)

        BaseLogic.update(answer, validated_data, user)

        if answer.question.type == models.Question.TYPE_TABLE:
            answer.create_answer_documents(documents)

        answer.refresh_from_db()
        return answer

    @staticmethod
    def set_file(validated_data):
        file_name = validated_data.get("file")
        file = models.File.objects.create(name=file_name)
        validated_data["file"] = file
        return validated_data


class SaveDefaultAnswerLogic(SaveAnswerLogic):
    @staticmethod
    def validate_for_save(
        data: dict, user: BaseUser, answer: models.Answer = None, origin: bool = False
    ) -> dict:
        if data["question"].type in [
            models.Question.TYPE_FILE,
            models.Question.TYPE_STATIC,
            models.Question.TYPE_DYNAMIC_CHOICE,
            models.Question.TYPE_DYNAMIC_MULTIPLE_CHOICE,
            models.Question.TYPE_CALCULATED_FLOAT,
            models.Question.TYPE_ACTION_BUTTON,
        ]:
            raise ValidationError(
                f'Can\'t save default_answer for question of type "{data["question"].type}"'
            )

        data["document"] = None  # send None as document for validation
        return SaveAnswerLogic.validate_for_save(data, user, answer, origin)

    @staticmethod
    @transaction.atomic
    def create(validated_data: dict, user: Optional[BaseUser] = None) -> models.Answer:
        answer = SaveAnswerLogic.create(validated_data, user)
        answer.question.default_answer = answer
        answer.question.save()

        return answer

    @staticmethod
    @transaction.atomic
    def update(
        answer, validated_data, user: Optional[BaseUser] = None
    ) -> models.Answer:
        return SaveAnswerLogic.update(answer, validated_data, user)


class SaveDocumentLogic:
    @staticmethod
    def _set_default_answers_for_form(form, document, user):
        """
        Set answers for questions with a default_answer.

        This method sets the default_answers on a new document. It recusively travels
        through the form and it's subforms, looking for questions that have a
        default_answer.
        For adding the answers, it does one query per form and one query per
        row_document.
        """
        answers = []
        for question in form.questions.filter(
            Q(default_answer__isnull=False) | Q(type=models.Question.TYPE_FORM)
        ).iterator():
            if question.type == models.Question.TYPE_FORM:
                if question.sub_form is not None:
                    document = SaveDocumentLogic._set_default_answers_for_form(
                        question.sub_form, document, user
                    )
                continue
            answers.append(
                question.default_answer.copy(document_family=document.family, user=user)
            )
        document.answers.add(*answers)
        return document

    @staticmethod
    @transaction.atomic
    def create(
        validated_data: dict, user: Optional[BaseUser] = None
    ) -> models.Document:
        document = BaseLogic.create(models.Document, validated_data, user=user)

        document = SaveDocumentLogic._set_default_answers_for_form(
            document.form, document, user
        )
        return document

    @staticmethod
    @transaction.atomic
    def update(document, validated_data, user: Optional[BaseUser] = None):
        BaseLogic.update(document, validated_data, user)


class CopyFormLogic:
    @staticmethod
    def copy(validated_data: dict, user: Optional[BaseUser] = None):
        source = validated_data["source"]
        validated_data["meta"] = dict(source.meta)
        form = BaseLogic.create(models.Form, validated_data, user=user)

        for form_question in models.FormQuestion.objects.filter(form=source):
            BaseLogic.create(
                models.FormQuestion,
                {
                    "form": form,
                    "sort": form_question.sort,
                    "question": form_question.question,
                },
                user,
            )

        return form



import pytest
from minio import Minio, S3Error

from ...caluma_form.models import File, Question


def test_delete_file_answer(
    db,
    question_factory,
    answer_factory,
    form_question_factory,
    file_factory,
    form,
    document,
    mocker,
):
    file_question = question_factory(type=Question.TYPE_FILE)
    form_question_factory(question=file_question, form=form)
    answer = answer_factory(
        question=file_question, value=None, document=document, file=file_factory()
    )
    mocker.patch.object(Minio, "copy_object")
    mocker.patch.object(Minio, "remove_object")
    answer.delete()
    with pytest.raises(File.DoesNotExist):
        answer.file.refresh_from_db()


def test_update_file(db, file_factory, mocker):
    file = file_factory()
    mocker.patch.object(Minio, "copy_object")
    mocker.patch.object(Minio, "remove_object")
    file.name = "something else"
    file.save()
    Minio.copy_object.assert_called()
    Minio.remove_object.assert_called()


@pytest.mark.parametrize("should_raise", [False, True])
def test_missing_file(db, file, mocker, should_raise):
    mocker.patch.object(
        Minio,
        "copy_object",
        side_effect=S3Error(
            code="SomeOtherError" if should_raise else "NoSuchKey",
            message="",
            resource="test_object",
            request_id="",
            host_id="",
            response=None,
        ),
    )

    if should_raise:
        with pytest.raises(S3Error):
            file.copy()
    else:
        file.copy()



# Generated by Django 2.2.3 on 2019-07-29 09:08

import django.contrib.postgres.indexes
from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [("caluma_form", "0021_auto_20190708_0905")]

    operations = [
        migrations.RemoveIndex(
            model_name="answer", name="form_answer_value_70b2d0_idx"
        ),
        migrations.AddIndex(
            model_name="answer",
            index=django.contrib.postgres.indexes.GinIndex(
                fields=["meta", "value"], name="form_answer_meta_98edaa_gin"
            ),
        ),
        migrations.AddIndex(
            model_name="document",
            index=django.contrib.postgres.indexes.GinIndex(
                fields=["meta"], name="form_docume_meta_333f0b_gin"
            ),
        ),
        migrations.AddIndex(
            model_name="form",
            index=django.contrib.postgres.indexes.GinIndex(
                fields=["meta"], name="form_form_meta_a2eed1_gin"
            ),
        ),
        migrations.AddIndex(
            model_name="option",
            index=django.contrib.postgres.indexes.GinIndex(
                fields=["meta"], name="form_option_meta_c09f31_gin"
            ),
        ),
        migrations.AddIndex(
            model_name="question",
            index=django.contrib.postgres.indexes.GinIndex(
                fields=["meta"], name="form_questi_meta_ea1ba0_gin"
            ),
        ),
    ]



import pytest

from ...caluma_core.models import UUIDModel
from ...caluma_core.tests.fake_model import get_fake_model
from ...caluma_core.types import DjangoObjectType
from ..visibilities import Authenticated, CreatedByGroup


@pytest.mark.parametrize("info_fixture,size", [("info", 0), ("admin_info", 1)])
def test_authenticated_visibility(db, info_fixture, size, request, history_mock):
    info = request.getfixturevalue(info_fixture)
    FakeModel = get_fake_model(model_base=UUIDModel)

    class CustomNode(DjangoObjectType):
        class Meta:
            model = FakeModel

    FakeModel.objects.create()

    queryset = Authenticated().filter_queryset(CustomNode, FakeModel.objects, info)
    assert queryset.count() == size


@pytest.mark.parametrize("group,size", [("unknown", 0), ("group", 1)])
def test_created_by_group_visibility(
    db, admin_info, group, admin_user, size, history_mock
):
    admin_user.groups = [group]
    FakeModel = get_fake_model(model_base=UUIDModel)

    class CustomNode(DjangoObjectType):
        class Meta:
            model = FakeModel

    FakeModel.objects.create(created_by_group="group")

    queryset = CreatedByGroup().filter_queryset(
        CustomNode, FakeModel.objects, admin_info
    )
    assert queryset.count() == size



from datetime import timedelta
from functools import wraps
from logging import getLogger

import minio
import urllib3
from django.conf import settings
from minio.commonconfig import CopySource
from minio.error import S3Error

log = getLogger(__name__)


class Retry(BaseException):
    pass


def _retry_on_missing_bucket(fn):
    """Create missing bucket if needed (decorator).

    If enabled in the settings, try to create the bucket if it
    doesn't exist yet, then retry.
    """

    @wraps(fn)
    def wrapper(self, *args, **kwargs):
        try:
            return fn(self, *args, **kwargs)
        except S3Error as exc:
            if (
                exc.code == "NoSuchBucket"
                and settings.MINIO_STORAGE_AUTO_CREATE_MEDIA_BUCKET
            ):
                log.warning(
                    b"Minio bucket '{self.bucket}' missing, trying to create it"
                )
                self.client.make_bucket(self.bucket)
                return fn(self, *args, **kwargs)
            raise

    return wrapper


class Minio:
    def __init__(self):
        endpoint = settings.MINIO_STORAGE_ENDPOINT
        access_key = settings.MINIO_STORAGE_ACCESS_KEY
        secret_key = settings.MINIO_STORAGE_SECRET_KEY
        secure = settings.MINIO_STORAGE_USE_HTTPS

        minio_http_client = None
        if settings.DEBUG and settings.MINIO_DISABLE_CERT_CHECKS:
            # This is a copy of what the minio client does internally,
            # just with the cert checking disabled.
            minio_http_client = urllib3.PoolManager(
                timeout=urllib3.Timeout.DEFAULT_TIMEOUT,
                maxsize=10,
                cert_reqs="CERT_NONE",
                ca_certs=None,
                retries=urllib3.Retry(
                    total=5, backoff_factor=0.2, status_forcelist=[500, 502, 503, 504]
                ),
            )
        self.client = minio.Minio(
            endpoint,
            access_key=access_key,
            secret_key=secret_key,
            secure=secure,
            http_client=minio_http_client,
        )
        self.bucket = settings.MINIO_STORAGE_MEDIA_BUCKET_NAME

    @_retry_on_missing_bucket
    def stat_object(self, object_name):
        """
        Get stat of object in bucket.

        :param object_name: str
        :return: stat response if successful, otherwise None
        """
        try:
            return self.client.stat_object(self.bucket, object_name)
        except S3Error as exc:
            log.error(f"Minio error, cannot stat object: {exc.code}")
            return None

    @_retry_on_missing_bucket
    def download_url(self, object_name):
        return self.client.presigned_get_object(
            self.bucket,
            object_name,
            timedelta(minutes=settings.MINIO_PRESIGNED_TTL_MINUTES),
        )

    @_retry_on_missing_bucket
    def upload_url(self, object_name):
        return self.client.presigned_put_object(
            self.bucket,
            object_name,
            timedelta(minutes=settings.MINIO_PRESIGNED_TTL_MINUTES),
        )

    @_retry_on_missing_bucket
    def remove_object(self, object_name):
        self.client.remove_object(self.bucket, object_name)

    @_retry_on_missing_bucket
    def copy_object(self, object_name, new_object_name):
        self.client.copy_object(
            self.bucket,
            new_object_name,
            CopySource(self.bucket, object_name),
        )

    @_retry_on_missing_bucket
    def move_object(self, object_name, new_object_name):
        self.copy_object(object_name, new_object_name)
        self.remove_object(object_name)


if settings.MEDIA_STORAGE_SERVICE == "minio":
    client = Minio()
else:  # pragma: no cover
    client = None
    raise NotImplementedError(
        f"Storage service {settings.MEDIA_STORAGE_SERVICE} is not implemented!"
    )



import json
import math
import numbers
from functools import partial
from itertools import count
from logging import getLogger

import pyjexl
from pyjexl.analysis import JEXLAnalyzer, ValidatingAnalyzer
from pyjexl.exceptions import ParseError
from pyjexl.parser import ArrayLiteral, Literal, ObjectLiteral
from rest_framework import exceptions

log = getLogger(__name__)


class Cache:
    """Custom cache.

    For JEXL expressions, we cannot use django's cache infrastructure, as the
    cached objects are pickled. This won't work for parsed JEXL expressions, as
    they contain lambdas etc.
    """

    def __init__(self, max_size=2000, evict_to=1500):
        self.max_size = max_size
        self.evict_to = evict_to

        self._cache = {}
        self._mru = {}
        self._mru_count = count()

    def get_or_set(self, key, default):
        if key in self._cache:
            self._mru[key] = next(self._mru_count)
            return self._cache[key]

        ret = self._cache[key] = default()
        self._mru[key] = next(self._mru_count)

        if len(self._mru) > self.max_size:
            self._evict()

        return ret

    def _evict(self):
        to_purge = list(sorted(self._cache.keys(), key=lambda key: self._mru[key]))
        # to_purge contains all keys, but we only want to evict the oldest
        # ones
        num_to_evict = len(to_purge) - self.evict_to

        for key in to_purge[:num_to_evict]:
            del self._cache[key]
            del self._mru[key]


class JexlValidator(object):
    def __init__(self, jexl):
        self.jexl = jexl

    def __call__(self, value):
        errors = list(self.jexl.validate(value))
        if errors:
            raise exceptions.ValidationError(errors)


class JEXL(pyjexl.JEXL):
    expr_cache = Cache()

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.add_transform("debug", self._debug_transform)
        self._expr_stack = []

        self.add_transform("mapby", self._mapby_transform)
        self.add_transform("stringify", lambda obj: json.dumps(obj))
        self.add_binary_operator(
            "intersects", 20, lambda left, right: any(x in right for x in left)
        )

        def _handle_error(func, subject, *args):
            try:
                if not args:
                    return func(subject)
                return func(subject, *args)
            except (TypeError, ValueError, ZeroDivisionError):
                return None

        def _handle_error_with_filter(func, subject, *args):
            return _handle_error(
                func,
                [
                    s
                    for s in subject
                    if isinstance(s, numbers.Number) and not math.isnan(s)
                ],
                *args,
            )

        # list operations
        self.add_transform("min", partial(_handle_error_with_filter, min))
        self.add_transform("max", partial(_handle_error_with_filter, max))
        self.add_transform("sum", partial(_handle_error_with_filter, sum))
        self.add_transform(
            "avg",
            partial(_handle_error_with_filter, lambda arr: sum(arr) / len(arr)),
        )
        # float operations
        self.add_transform("ceil", partial(_handle_error, math.ceil))
        self.add_transform("floor", partial(_handle_error, math.floor))
        self.add_transform(
            "round",
            partial(
                _handle_error,
                self._round_compat,
            ),
        )

    def _round_compat(self, num, ndigits=0):
        power = 10 ** ndigits
        return float(math.floor((num * power) + 0.5) / power)

    def parse(self, expression):
        parsed_expression = self.expr_cache.get_or_set(
            expression, lambda: super(JEXL, self).parse(expression)
        )
        return parsed_expression

    def validate(self, expression, ValidatingAnalyzerClass=ValidatingAnalyzer):
        try:
            for res in self.analyze(expression, ValidatingAnalyzerClass):
                yield res
        except ParseError as err:
            yield str(err)

    def _debug_transform(self, value, label=None):
        expr = self._expr_stack[-1]

        log.info(
            f"JEXL debug ({label}): value = `{value}`"
            if label
            else f"JEXL debug: in expression `{expr}`, value = `{value}`"
        )
        return value

    def _mapby_transform(self, arr, *keys):
        if not isinstance(arr, list):
            return None

        return [
            [obj.get(key) for key in keys] if len(keys) > 1 else obj.get(keys[0])
            for obj in arr
        ]

    def evaluate(self, expression, context=None):
        self._expr_stack.append(expression)
        try:
            return super().evaluate(expression, context)
        finally:
            self._expr_stack.pop()


class CalumaAnalyzer(JEXLAnalyzer):
    """Analyzer visiting Object and ArrayLiterals.

    TODO: Upstream this.
    """

    def __init__(self, config, transforms=None):
        self.transforms = transforms if transforms else []
        super().__init__(config)

    def generic_visit(self, expression):
        for child in expression.children:
            assert child is not None
            for c in self.visit(child):
                yield c

        if isinstance(expression, ArrayLiteral):
            for child in expression.value:
                assert child is not None
                for c in self.visit(child):
                    yield c

        elif isinstance(expression, ObjectLiteral):
            for child in expression.value.values():
                assert child is not None
                for c in self.visit(child):
                    yield c


class ExtractTransformSubjectAnalyzer(CalumaAnalyzer):
    """
    Extract all referenced subjects of given transforms.

    If no transforms are given all references of all transforms will be extracted.
    """

    def visit_Transform(self, transform):
        if not self.transforms or transform.name in self.transforms:
            # can only extract subject's value if subject is a Literal
            if not isinstance(transform.subject, type(transform)):
                yield transform.subject.value

        yield from self.generic_visit(transform)


class ExtractTransformArgumentAnalyzer(CalumaAnalyzer):
    """
    Extract all referenced arguments of given transforms.

    If no transforms are given all references of all transforms will be extracted.
    """

    def visit_Transform(self, transform):
        if not self.transforms or transform.name in self.transforms:
            # "mapby" is the only transform that accepts args, those can hold
            # references. We need a "lookahead" to peek if the subject is an
            # answer transform in this case we assume the arguments are
            # references (as opposed to simple literals)
            if (
                transform.name == "mapby"
                and transform.subject
                and transform.subject.name == "answer"
                and len(transform.args)
            ):
                for arg in transform.args:
                    if isinstance(arg, Literal):
                        yield arg.value

        yield from self.generic_visit(transform)


class ExtractTransformSubjectAndArgumentsAnalyzer(CalumaAnalyzer):
    """
    Extract all referenced subjects and arguments of a given transforms.

    If no transforms are given all references of all transforms will be extracted.
    """

    def visit_Transform(self, transform):
        if not self.transforms or transform.name in self.transforms:
            if not isinstance(transform.subject, type(transform)):
                yield (transform.subject.value, transform.args)

        yield from self.generic_visit(transform)



# Generated by Django 2.2.22 on 2021-11-03 08:52

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("caluma_form", "0040_add_modified_by_user_group"),
    ]

    operations = [
        migrations.AlterField(
            model_name="historicalquestion",
            name="type",
            field=models.CharField(
                choices=[
                    ("multiple_choice", "multiple_choice"),
                    ("integer", "integer"),
                    ("float", "float"),
                    ("date", "date"),
                    ("choice", "choice"),
                    ("textarea", "textarea"),
                    ("text", "text"),
                    ("table", "table"),
                    ("form", "form"),
                    ("file", "file"),
                    ("dynamic_choice", "dynamic_choice"),
                    ("dynamic_multiple_choice", "dynamic_multiple_choice"),
                    ("static", "static"),
                    ("calculated_float", "calculated_float"),
                    ("action_button", "action_button"),
                ],
                max_length=23,
            ),
        ),
        migrations.AlterField(
            model_name="question",
            name="type",
            field=models.CharField(
                choices=[
                    ("multiple_choice", "multiple_choice"),
                    ("integer", "integer"),
                    ("float", "float"),
                    ("date", "date"),
                    ("choice", "choice"),
                    ("textarea", "textarea"),
                    ("text", "text"),
                    ("table", "table"),
                    ("form", "form"),
                    ("file", "file"),
                    ("dynamic_choice", "dynamic_choice"),
                    ("dynamic_multiple_choice", "dynamic_multiple_choice"),
                    ("static", "static"),
                    ("calculated_float", "calculated_float"),
                    ("action_button", "action_button"),
                ],
                max_length=23,
            ),
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.11.17 on 2018-12-17 10:51
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [("caluma_workflow", "0002_auto_20181213_1334")]

    operations = [
        migrations.AlterField(
            model_name="task",
            name="type",
            field=models.CharField(
                choices=[
                    ("simple", "Task which can only be marked as completed."),
                    (
                        "complete_workflow_form",
                        "Task completing defined workflow form.",
                    ),
                ],
                max_length=50,
            ),
        )
    ]



import pytest

from ..dynamic_tasks import BaseDynamicTasks, register_dynamic_task
from ..jexl import FlowJexl, GroupJexl


class CustomDynamicTasks(BaseDynamicTasks):
    @register_dynamic_task("dynamic-task")
    def resolve_dynamic_task(
        self, case, user, prev_work_item, context
    ):  # pragma: no cover
        return ["task-slug", "task-slug1", "task-slug2"]


@pytest.fixture
def task_config(db, task_factory):
    FlowJexl.dynamic_tasks_classes = [CustomDynamicTasks]

    for slug in ["task-slug", "task-slug1", "task-slug2"]:
        task_factory(slug=slug)


@pytest.mark.parametrize(
    "expression,expected_tasks",
    [
        (
            "63 > 62 ? 'test'|task : ['test2', 'test3']|tasks",
            {"test", "test2", "test3"},
        ),
        ("['test2', 'test3']|tasks", {"test2", "test3"}),
        ("'test'|task", {"test"}),
    ],
)
def test_flow_extract_tasks(expression, expected_tasks):
    jexl = FlowJexl()

    assert expected_tasks == set(jexl.extract_tasks(expression))


@pytest.mark.parametrize(
    "expression,expected_errors",
    [
        # correct tasks
        ("'task-slug'|task", []),
        ("['task-slug1', 'task-slug2', 'dynamic-task']|tasks", []),
        # invalid subject type
        ("100|task", ["`100` is not a valid task slug."]),
        ("'task-slug'|tasks", ["`task-slug` is not a valid list of tasks."]),
        # not existing tasks
        (
            "'some-task'|task",
            [
                "The task `some-task` does not exist or is not registered as dynamic task"
            ],
        ),
        (
            "['some-task', 'some-other-task']|tasks",
            [
                "The task `some-task` does not exist or is not registered as dynamic task",
                "The task `some-other-task` does not exist or is not registered as dynamic task",
            ],
        ),
    ],
)
def test_flow_jexl_validate(expression, expected_errors, task_config):
    errors = sorted(FlowJexl().validate(expression))

    assert sorted(expected_errors) == errors


@pytest.mark.parametrize(
    "expression,num_errors",
    [
        # correct cases
        ("['group1']|groups", 0),
        # invalid subject type
        ("100|groups", 1),
        ("'task-slug'|groups", 1),
    ],
)
def test_group_jexl_validate(expression, num_errors):
    jexl = GroupJexl()
    assert len(list(jexl.validate(expression))) == num_errors



import pytest

from caluma.caluma_core.relay import extract_global_id
from caluma.caluma_core.tests import (
    extract_global_id_input_fields,
    extract_serializer_input_fields,
)

from .. import api, models, serializers


@pytest.mark.parametrize(
    "question__type,question__configuration,question__data_source,question__format_validators",
    [
        (models.Question.TYPE_INTEGER, {"max_value": 10, "min_value": 0}, None, []),
        (models.Question.TYPE_FLOAT, {"max_value": 1.0, "min_value": 0.0}, None, []),
        (models.Question.TYPE_FLOAT, {}, None, []),
        (models.Question.TYPE_DATE, {}, None, []),
        (models.Question.TYPE_TEXT, {"min_length": 10}, None, ["email"]),
        (models.Question.TYPE_TEXTAREA, {"max_length": 10}, None, []),
        (models.Question.TYPE_CHOICE, {}, None, []),
        (models.Question.TYPE_MULTIPLE_CHOICE, {}, None, []),
        (models.Question.TYPE_FORM, {}, None, []),
        (models.Question.TYPE_FILE, {}, None, []),
        (models.Question.TYPE_DYNAMIC_CHOICE, {}, "MyDataSource", []),
        (models.Question.TYPE_DYNAMIC_MULTIPLE_CHOICE, {}, "MyDataSource", []),
        (models.Question.TYPE_STATIC, {}, None, []),
        (models.Question.TYPE_CALCULATED_FLOAT, {}, None, []),
        (models.Question.TYPE_ACTION_BUTTON, {}, None, []),
    ],
)
def test_query_all_questions(
    schema_executor,
    db,
    snapshot,
    question,
    form,
    form_question_factory,
    question_option,
    data_source_settings,
):
    form_question_factory.create(form=form)

    query = """
        query AllQuestionsQuery($search: String, $forms: [ID]) {
          allQuestions(isArchived: false, search: $search, excludeForms: $forms) {
            totalCount
            edges {
              node {
                id
                __typename
                slug
                label
                meta
                infoText
                ... on TextQuestion {
                  minLength
                  maxLength
                  placeholder
                  formatValidators {
                    edges {
                      node {
                        slug
                        name
                        regex
                        errorMsg
                      }
                    }
                  }
                }
                ... on TextareaQuestion {
                  minLength
                  maxLength
                  placeholder
                  formatValidators {
                    edges {
                      node {
                        slug
                        name
                        regex
                        errorMsg
                      }
                    }
                  }
                }
                ... on FloatQuestion {
                  floatMinValue: minValue
                  floatMaxValue: maxValue
                  placeholder
                }
                ... on IntegerQuestion {
                  integerMinValue: minValue
                  integerMaxValue: maxValue
                  placeholder
                }
                ... on MultipleChoiceQuestion {
                  options {
                    totalCount
                    edges {
                      node {
                        slug
                      }
                    }
                  }
                }
                ... on ChoiceQuestion {
                  options {
                    totalCount
                    edges {
                      node {
                        slug
                      }
                    }
                  }
                }
                ... on DynamicMultipleChoiceQuestion {
                  options {
                    edges {
                      node {
                        slug
                        label
                      }
                    }
                  }
                }
                ... on DynamicChoiceQuestion {
                  options {
                    edges {
                      node {
                        slug
                        label
                      }
                    }
                  }
                }
                ... on FormQuestion {
                  subForm {
                    slug
                  }
                }
                ... on StaticQuestion {
                  staticContent
                }
                ... on CalculatedFloatQuestion {
                  calcExpression
                }
              }
            }
          }
        }
    """

    result = schema_executor(
        query,
        variable_values={
            "search": question.label,
            "forms": [extract_global_id_input_fields(form)["id"]],
        },
    )

    assert not result.errors
    snapshot.assert_match(result.data)


@pytest.mark.parametrize("question__meta", [{"meta": "set"}])
def test_copy_question(db, question, question_option_factory, schema_executor):
    def sorted_option_list(q):
        return sorted(
            list(
                models.QuestionOption.objects.filter(question=q).values_list(
                    "option", flat=True
                )
            )
        )

    question_option_factory.create_batch(5, question=question)
    query = """
        mutation CopyQuestion($input: CopyQuestionInput!) {
          copyQuestion(input: $input) {
            question {
              slug
            }
            clientMutationId
          }
        }
    """

    inp = {
        "input": {
            "source": question.pk,
            "slug": "new-question",
            "label": "Test Question",
        }
    }
    result = schema_executor(query, variable_values=inp)

    assert not result.errors

    question_slug = result.data["copyQuestion"]["question"]["slug"]
    assert question_slug == "new-question"
    new_question = models.Question.objects.get(pk=question_slug)
    assert new_question.label == "Test Question"
    assert new_question.meta == question.meta
    assert new_question.type == question.type
    assert new_question.configuration == question.configuration
    assert new_question.row_form == question.row_form
    assert new_question.is_hidden == question.is_hidden
    assert new_question.is_required == question.is_required
    assert new_question.source == question
    assert sorted_option_list(new_question) == sorted_option_list(question)


@pytest.mark.parametrize(
    "mutation",
    [
        "SaveTextQuestion",
        "SaveTextareaQuestion",
        "SaveIntegerQuestion",
        "SaveFloatQuestion",
        "SaveDateQuestion",
        "SaveFileQuestion",
        "SaveCalculatedFloatQuestion",
    ],
)
@pytest.mark.parametrize(
    "question__is_required,success", (("true", True), ("true|invalid", False))
)
def test_save_question(db, snapshot, question, mutation, schema_executor, success):
    mutation_func = mutation[0].lower() + mutation[1:]
    query = f"""
        mutation {mutation}($input: {mutation}Input!) {{
          {mutation_func}(input: $input) {{
            question {{
              id
              slug
              label
              meta
              __typename
            }}
            clientMutationId
          }}
        }}
    """

    inp = {
        "input": extract_serializer_input_fields(
            serializers.SaveQuestionSerializer, question
        )
    }
    result = schema_executor(query, variable_values=inp)

    assert not bool(result.errors) == success
    if success:
        snapshot.assert_match(result.data)


@pytest.mark.parametrize(
    "question__type,question__configuration,question__format_validators,answer__value,success",
    [
        (models.Question.TYPE_TEXT, {"max_length": 10}, [], "foo", True),
        (models.Question.TYPE_TEXT, {"max_length": 10}, ["email"], "foo", True),
        (
            models.Question.TYPE_TEXT,
            {"max_length": 10},
            ["notavalidator"],
            "foo",
            False,
        ),
    ],
)
def test_save_text_question(db, question, schema_executor, answer, success):
    query = """
        mutation SaveTextQuestion($input: SaveTextQuestionInput!) {
          saveTextQuestion(input: $input) {
            question {
              id
              slug
              label
              meta
              __typename
              ... on TextQuestion {
                maxLength
                formatValidators {
                  edges {
                    node {
                      slug
                      name
                      regex
                      errorMsg
                    }
                  }
                }
                defaultAnswer {
                  value
                }
              }
            }
            clientMutationId
          }
        }
    """

    question.default_answer = answer
    question.save()

    inp = {
        "input": extract_serializer_input_fields(
            serializers.SaveTextQuestionSerializer, question
        )
    }
    result = schema_executor(query, variable_values=inp)
    assert not bool(result.errors) == success
    if success:
        assert result.data["saveTextQuestion"]["question"]["maxLength"] == 10
        if question.format_validators:
            assert (
                result.data["saveTextQuestion"]["question"]["formatValidators"][
                    "edges"
                ][0]["node"]["slug"]
                == "email"
            )
        assert (
            result.data["saveTextQuestion"]["question"]["defaultAnswer"]["value"]
            == "foo"
        )


@pytest.mark.parametrize(
    "question__type,question__configuration,answer__value",
    [(models.Question.TYPE_TEXTAREA, {"max_length": 10}, "foo")],
)
def test_save_textarea_question(db, question, answer, schema_executor):
    query = """
        mutation SaveTextareaQuestion($input: SaveTextareaQuestionInput!) {
          saveTextareaQuestion(input: $input) {
            question {
              id
              slug
              label
              meta
              __typename
              ... on TextareaQuestion {
                maxLength
                defaultAnswer {
                  value
                }
              }
            }
            clientMutationId
          }
        }
    """

    question.default_answer = answer
    question.save()

    inp = {
        "input": extract_serializer_input_fields(
            serializers.SaveTextareaQuestionSerializer, question
        )
    }
    result = schema_executor(query, variable_values=inp)
    assert not result.errors
    assert result.data["saveTextareaQuestion"]["question"]["maxLength"] == 10
    assert (
        result.data["saveTextareaQuestion"]["question"]["defaultAnswer"]["value"]
        == "foo"
    )


@pytest.mark.parametrize(
    "question__type,question__configuration,answer__value,success",
    [
        (models.Question.TYPE_FLOAT, {"max_value": 10.0, "min_value": 0.0}, 0.3, True),
        (models.Question.TYPE_FLOAT, {"max_value": 1.0, "min_value": 10.0}, 0.3, False),
    ],
)
def test_save_float_question(db, snapshot, question, schema_executor, answer, success):
    query = """
        mutation SaveFloatQuestion($input: SaveFloatQuestionInput!) {
          saveFloatQuestion(input: $input) {
            question {
              id
              slug
              label
              meta
              __typename
              ... on FloatQuestion {
                minValue
                maxValue
                defaultAnswer {
                  value
                }
              }
            }
            clientMutationId
          }
        }
    """

    question.default_answer = answer
    question.save()

    inp = {
        "input": extract_serializer_input_fields(
            serializers.SaveFloatQuestionSerializer, question
        )
    }
    result = schema_executor(query, variable_values=inp)
    assert not bool(result.errors) == success
    if success:
        snapshot.assert_match(result.data)
        assert (
            result.data["saveFloatQuestion"]["question"]["defaultAnswer"]["value"]
            == 0.3
        )


@pytest.mark.parametrize(
    "question__type,question__configuration,answer__value,success",
    [
        (models.Question.TYPE_INTEGER, {"max_value": 10, "min_value": 0}, 23, True),
        (models.Question.TYPE_INTEGER, {"max_value": 1, "min_value": 10}, 23, False),
    ],
)
def test_save_integer_question(
    db, snapshot, question, answer, success, schema_executor
):
    query = """
        mutation SaveIntegerQuestion($input: SaveIntegerQuestionInput!) {
          saveIntegerQuestion(input: $input) {
            question {
              id
              slug
              label
              meta
              __typename
              ... on IntegerQuestion {
                minValue
                maxValue
                defaultAnswer {
                  value
                }
              }
            }
            clientMutationId
          }
        }
    """

    question.default_answer = answer
    question.save()

    inp = {
        "input": extract_serializer_input_fields(
            serializers.SaveIntegerQuestionSerializer, question
        )
    }
    result = schema_executor(query, variable_values=inp)
    assert not bool(result.errors) == success
    if success:
        snapshot.assert_match(result.data)
        assert (
            result.data["saveIntegerQuestion"]["question"]["defaultAnswer"]["value"]
            == 23
        )


@pytest.mark.parametrize("question__type", [models.Question.TYPE_MULTIPLE_CHOICE])
def test_save_multiple_choice_question(
    db, snapshot, question, question_option_factory, answer_factory, schema_executor
):
    question_option_factory.create_batch(2, question=question)

    option_ids = question.options.order_by("-slug").values_list("slug", flat=True)

    question.default_answer = answer_factory(value=list(option_ids), question=question)
    question.save()

    query = """
        mutation SaveMultipleChoiceQuestion($input: SaveMultipleChoiceQuestionInput!) {
          saveMultipleChoiceQuestion(input: $input) {
            question {
              id
              slug
              label
              meta
              __typename
              ... on MultipleChoiceQuestion {
                options {
                  edges {
                    node {
                      slug
                      label
                    }
                  }
                }
                defaultAnswer {
                  value
                }
              }
            }
            clientMutationId
          }
        }
    """

    inp = {
        "input": extract_serializer_input_fields(
            serializers.SaveMultipleChoiceQuestionSerializer, question
        )
    }
    inp["input"]["options"] = option_ids
    result = schema_executor(query, variable_values=inp)
    assert not result.errors
    snapshot.assert_match(result.data)


@pytest.mark.parametrize("with_default", [True, False])
@pytest.mark.parametrize("question__type", [models.Question.TYPE_CHOICE])
def test_save_choice_question(
    db,
    snapshot,
    question,
    question_option,
    answer_factory,
    schema_executor,
    with_default,
):
    query = """
        mutation SaveChoiceQuestion($input: SaveChoiceQuestionInput!) {
          saveChoiceQuestion(input: $input) {
            question {
              id
              slug
              label
              meta
              __typename
              ... on ChoiceQuestion {
                options {
                  edges {
                    node {
                      slug
                      label
                    }
                  }
                }
                defaultAnswer {
                  value
                }
              }
            }
            clientMutationId
          }
        }
    """

    if with_default:
        question.default_answer = answer_factory(
            value=question_option.option.slug, question=question
        )
        question.save()

    inp = {
        "input": extract_serializer_input_fields(
            serializers.SaveChoiceQuestionSerializer, question
        )
    }
    if not with_default:
        question.delete()
    result = schema_executor(query, variable_values=inp)
    assert not result.errors
    snapshot.assert_match(result.data)


@pytest.mark.parametrize("delete", [True, False])
@pytest.mark.parametrize(
    "question__type",
    [models.Question.TYPE_DYNAMIC_CHOICE, models.Question.TYPE_DYNAMIC_MULTIPLE_CHOICE],
)
def test_save_dynamic_choice_question(
    db, snapshot, question, delete, schema_executor, data_source_settings
):
    query = """
        mutation SaveDynamicChoiceQuestion($input: SaveDynamicChoiceQuestionInput!) {
          saveDynamicChoiceQuestion(input: $input) {
            question {
              id
              slug
              label
              meta
              __typename
              ... on DynamicChoiceQuestion {
                options {
                  edges {
                    node {
                      slug
                      label
                    }
                  }
                }
              }
            }
            clientMutationId
          }
        }
    """

    inp = {
        "input": extract_serializer_input_fields(
            serializers.SaveDynamicChoiceQuestionSerializer, question
        )
    }
    if delete:
        question.delete()  # test creation
    result = schema_executor(query, variable_values=inp)
    assert not result.errors
    snapshot.assert_match(result.data)


@pytest.mark.parametrize("delete", [True, False])
@pytest.mark.parametrize(
    "question__type", [models.Question.TYPE_DYNAMIC_MULTIPLE_CHOICE]
)
def test_save_dynamic_multiple_choice_question(
    db,
    snapshot,
    question,
    delete,
    question_option_factory,
    schema_executor,
    data_source_settings,
):
    query = """
        mutation SaveDynamicMultipleChoiceQuestion($input: SaveDynamicMultipleChoiceQuestionInput!) {
          saveDynamicMultipleChoiceQuestion(input: $input) {
            question {
              id
              slug
              label
              meta
              __typename
              ... on DynamicMultipleChoiceQuestion {
                options {
                  edges {
                    node {
                      slug
                      label
                    }
                  }
                }
              }
            }
            clientMutationId
          }
        }
    """

    inp = {
        "input": extract_serializer_input_fields(
            serializers.SaveDynamicMultipleChoiceQuestionSerializer, question
        )
    }
    if delete:
        question.delete()  # test creation
    result = schema_executor(query, variable_values=inp)
    assert not result.errors
    snapshot.assert_match(result.data)


@pytest.mark.parametrize("question__type", [models.Question.TYPE_TABLE])
def test_save_table_question(db, snapshot, question, schema_executor):
    query = """
        mutation SaveTableQuestion($input: SaveTableQuestionInput!) {
          saveTableQuestion(input: $input) {
            question {
              id
              slug
              label
              meta
              __typename
              ... on TableQuestion {
                rowForm {
                  slug
                }
              }
            }
            clientMutationId
          }
        }
    """

    inp = {
        "input": extract_serializer_input_fields(
            serializers.SaveTableQuestionSerializer, question
        )
    }
    question.delete()  # test creation
    result = schema_executor(query, variable_values=inp)
    assert not result.errors
    snapshot.assert_match(result.data)


@pytest.mark.parametrize("question__type", [models.Question.TYPE_FORM])
def test_save_form_question(db, snapshot, question, question_option, schema_executor):
    query = """
        mutation SaveFormQuestion($input: SaveFormQuestionInput!) {
          saveFormQuestion(input: $input) {
            question {
              id
              slug
              label
              meta
              __typename
              ... on FormQuestion {
                subForm {
                  slug
                }
              }
            }
            clientMutationId
          }
        }
    """

    inp = {
        "input": extract_serializer_input_fields(
            serializers.SaveFormQuestionSerializer, question
        )
    }
    question.delete()  # test creation
    result = schema_executor(query, variable_values=inp)
    assert not result.errors
    snapshot.assert_match(result.data)


@pytest.mark.parametrize("question__type", [models.Question.TYPE_STATIC])
def test_save_static_question(db, snapshot, question, schema_executor):
    query = """
        mutation SaveStaticQuestion($input: SaveStaticQuestionInput!) {
          saveStaticQuestion(input: $input) {
            question {
              id
              slug
              label
              meta
              __typename
              ... on StaticQuestion {
                staticContent
              }
            }
            clientMutationId
          }
        }
    """

    inp = {
        "input": extract_serializer_input_fields(
            serializers.SaveStaticQuestionSerializer, question
        )
    }
    result = schema_executor(query, variable_values=inp)
    assert not bool(result.errors)
    snapshot.assert_match(result.data)


def test_all_questions_slug_filter(
    db,
    schema_executor,
    question_option_factory,
    question_factory,
    form_factory,
    form_question_factory,
    option_factory,
):
    query = """
        query FilteredQuestions {
          allQuestions {
            edges {
              node {
                slug
              }
            }
          }
        }
    """
    questions = question_factory.create_batch(4)

    forms = form_factory.create_batch(2)

    for question in questions:
        for form in forms:
            question.forms.through.objects.create(form=form, question=question)

    num_questions = models.Question.objects.all().count()
    result = schema_executor(query)

    assert not result.errors
    assert len(result.data["allQuestions"]["edges"]) == num_questions


def test_calc_dependents(db, question_factory):
    dep1, dep2, dep3 = question_factory.create_batch(3)

    assert not any([dep1.calc_dependents, dep2.calc_dependents, dep3.calc_dependents])

    calc_q = question_factory(
        type=models.Question.TYPE_CALCULATED_FLOAT,
        calc_expression=f"['{dep1.slug}'|answer, '{dep2.slug}'|answer]|sum",
    )
    other_calc_q = question_factory(
        type=models.Question.TYPE_CALCULATED_FLOAT,
        calc_expression=f"'{dep1.slug}'|answer + '{dep2.slug}'|answer * '{dep3.slug}'|answer",
    )
    dep1.refresh_from_db()
    dep2.refresh_from_db()
    dep3.refresh_from_db()

    def slug_in_deps(slug, deps):
        return [slug in dependents for dependents in [q.calc_dependents for q in deps]]

    assert all(slug_in_deps(calc_q.slug, [dep1, dep2]))
    assert all(slug_in_deps(other_calc_q.slug, [dep1, dep2, dep3]))

    calc_q.calc_expression = f"'{dep2.slug}'|answer + '{dep3.slug}'|answer"
    calc_q.save()
    dep1.refresh_from_db()
    dep2.refresh_from_db()
    dep3.refresh_from_db()

    assert dep1.calc_dependents == [other_calc_q.slug]
    assert all(slug_in_deps(calc_q.slug, [dep2, dep3]))
    assert all(slug_in_deps(other_calc_q.slug, [dep1, dep2, dep3]))

    calc_q.delete()

    dep1.refresh_from_db()
    dep2.refresh_from_db()
    dep3.refresh_from_db()
    assert not any(slug_in_deps(calc_q.slug, [dep1, dep2, dep3]))
    assert all(slug_in_deps(other_calc_q.slug, [dep1, dep2, dep3]))


@pytest.mark.parametrize("question__slug", ["dep-question"])
@pytest.mark.parametrize(
    "question__type,answer_value,calc_expression,expected",
    [
        (models.Question.TYPE_FLOAT, 1.5, "'dep-question'|answer * 3", 4.5),
        (models.Question.TYPE_FLOAT, 0, "'dep-question'|answer * 3", 0),
        (models.Question.TYPE_INTEGER, 100, "'dep-question'|answer", 100.0),
        (models.Question.TYPE_INTEGER, 100, "'dep-question'|answer / 10 * 99", 990.0),
    ],
)
def test_calculated_question(
    db,
    schema_executor,
    form_question,
    question_factory,
    answer_value,
    calc_expression,
    expected,
):
    form = form_question.form
    question = form_question.question

    question_factory(
        slug="not_in_form",
        type=models.Question.TYPE_CALCULATED_FLOAT,
        calc_expression=f"'{question.slug}'|answer",
    )

    query = """
        mutation SaveCalculatedQuestion($input: SaveCalculatedFloatQuestionInput!) {
          saveCalculatedFloatQuestion (input: $input) {
            question {
              slug
             __typename
             ... on CalculatedFloatQuestion {
               calcExpression
             }
            }
          }
        }
    """

    inp = {
        "input": {
            "slug": "calc-question",
            "label": "Calculated Float Question",
            "calcExpression": calc_expression,
        }
    }
    result = schema_executor(query, variable_values=inp)

    assert not result.errors

    question.refresh_from_db()
    assert set(question.calc_dependents) == set(["calc-question", "not_in_form"])

    calc_question = models.Question.objects.get(slug="calc-question")
    form.questions.add(calc_question)

    query = """
        mutation SaveDocument($input: SaveDocumentInput!) {
          saveDocument (input: $input) {
            document {
              id
              answers {
                edges {
                  node {
                    ... on FloatAnswer {
                      value
                    }
                  }
                }
              }
            }
          }
        }
    """

    inp = {"input": {"form": form.slug}}
    result = schema_executor(query, variable_values=inp)
    assert not result.errors

    calc_answer = calc_question.answers.first()
    assert calc_answer
    assert calc_answer.value is None

    q_type = question.type.capitalize()
    query = f"""
        mutation SaveDocument{q_type}Answer($input: SaveDocument{q_type}AnswerInput!) {{
          saveDocument{q_type}Answer (input: $input) {{
            answer {{
              id
            }}
          }}
        }}
    """

    inp = {
        "input": {
            "document": extract_global_id(
                result.data["saveDocument"]["document"]["id"]
            ),
            "question": question.slug,
            "value": answer_value,
        }
    }
    result = schema_executor(query, variable_values=inp)
    assert not result.errors

    calc_answer = calc_question.answers.filter().first()
    assert calc_answer
    assert calc_answer.value == expected


@pytest.mark.parametrize(
    "expr,expected,calc_deps",
    [
        ("'sub_question'|answer", 100.0, ["sub_question"]),
        ("'table'|answer|mapby('column')[0]", 99.99, ["table", "column"]),
        ("'other'|answer", 3.0, ["other"]),
        (
            "'sub_question'|answer && 'table'|answer|mapby('column', 'column2') ? 'other'|answer: -1",
            3.0,
            ["sub_question", "table", "column", "column2", "other"],
        ),
    ],
)
def test_nested_calculated_question(
    db,
    schema_executor,
    form_and_document,
    form_question_factory,
    question_factory,
    expr,
    expected,
    calc_deps,
):
    form, document, questions, answers_dict = form_and_document(True, True)

    sub_question = questions["sub_question"]
    sub_question.type = models.Question.TYPE_INTEGER
    sub_question.save()
    sub_question_a = answers_dict["sub_question"]
    sub_question_a.value = 100
    sub_question_a.save()

    column_a = answers_dict["column"]
    column_a.value = 99.99
    column_a.save()

    questions["other"] = form_question_factory(
        form=form, question__slug="other", question__type=models.Question.TYPE_INTEGER
    ).question

    questions["column2"] = question_factory(
        slug="column2", is_required="false", is_hidden="false"
    )
    form_question_factory(
        form=questions["table"].row_form, question=questions["column2"]
    )

    api.save_answer(question=questions["other"], document=document, value=3)

    questions["calc"] = form_question_factory(
        form=form,
        question__slug="calc",
        question__type=models.Question.TYPE_CALCULATED_FLOAT,
        question__calc_expression=expr,
    )

    calc_ans = document.answers.get(question_id="calc")
    assert calc_ans.value == expected

    for dep in calc_deps:
        questions[dep].refresh_from_db()
        assert questions[dep].calc_dependents == ["calc"]


@pytest.mark.parametrize("question__slug", ["simple-question"])
@pytest.mark.parametrize(
    "calc_expression,has_error",
    [
        ("1.0", False),
        ("'simple-question'|answer", False),
        ("'other-calc-question'|answer + 10", True),
    ],
)
def test_recursive_calculated_question(
    db, schema_executor, form_question, question_factory, calc_expression, has_error
):
    form = form_question.form

    other_q = question_factory(
        slug="other-calc-question",
        type=models.Question.TYPE_CALCULATED_FLOAT,
        calc_expression="0.0",
    )
    form.questions.add(other_q)

    query = """
        mutation SaveCalculatedQuestion($input: SaveCalculatedFloatQuestionInput!) {
          saveCalculatedFloatQuestion (input: $input) {
            question {
              slug
             __typename
             ... on CalculatedFloatQuestion {
               calcExpression
             }
            }
          }
        }
    """

    inp = {
        "input": {
            "slug": "calc-question",
            "label": "Calculated Float Question",
            "calcExpression": calc_expression,
        }
    }
    result = schema_executor(query, variable_values=inp)

    assert has_error == bool(result.errors)


def test_calculated_question_update_calc_expr(
    db, schema_executor, form_and_document, form_question_factory
):
    form, document, questions_dict, answers_dict = form_and_document(True, True)

    sub_question = questions_dict["sub_question"]
    sub_question.type = models.Question.TYPE_INTEGER
    sub_question.save()
    sub_question_a = answers_dict["sub_question"]
    sub_question_a.value = 100
    sub_question_a.save()

    calc_question = form_question_factory(
        form=form,
        question__slug="calc_question",
        question__type=models.Question.TYPE_CALCULATED_FLOAT,
        question__calc_expression="'sub_question'|answer + 1",
    ).question

    calc_ans = document.answers.get(question_id="calc_question")
    assert calc_ans.value == 101

    calc_question.calc_expression = "'sub_question'|answer -1"
    calc_question.save()
    calc_ans.refresh_from_db()
    assert calc_ans.value == 99


def test_calculated_question_answer_document(
    db,
    schema_executor,
    form_and_document,
    form_question_factory,
    answer_factory,
    document_factory,
):
    form, document, questions_dict, answers_dict = form_and_document(True, True)

    table = questions_dict["table"]
    table_a = answers_dict["table"]
    row_form = table.row_form
    column = questions_dict["column"]
    column.type = models.Question.TYPE_INTEGER
    column.save()

    column_a1 = column.answers.get()
    column_a1.value = 100
    column_a1.save()

    form_question_factory(
        form=form,
        question__slug="calc_question",
        question__type=models.Question.TYPE_CALCULATED_FLOAT,
        question__calc_expression="'table'|answer|mapby('column')[0] + 'table'|answer|mapby('column')[1]",
    ).question

    calc_ans = document.answers.get(question_id="calc_question")
    assert calc_ans.value is None

    # adding another row will make make the expression valid
    row_doc = document_factory(form=row_form, family=document)
    column_a2 = answer_factory(document=row_doc, question_id=column.slug, value=200)
    table_a.documents.add(row_doc)

    calc_ans.refresh_from_db()
    assert calc_ans.value == 300

    column_a2.value = 100
    column_a2.save()
    calc_ans.refresh_from_db()
    column.refresh_from_db()
    assert column.calc_dependents == ["calc_question"]
    assert calc_ans.value == 200

    # removing the row will make it invalid again
    table_a.documents.remove(row_doc)
    row_doc.delete()
    calc_ans.refresh_from_db()
    assert calc_ans.value is None


@pytest.mark.parametrize("question__type", [models.Question.TYPE_ACTION_BUTTON])
def test_save_action_button_question(db, snapshot, question, schema_executor):
    query = """
        mutation SaveActionButtonQuestion($input: SaveActionButtonQuestionInput!) {
          saveActionButtonQuestion(input: $input) {
            question {
              id
              slug
              label
              meta
              __typename
              ... on ActionButtonQuestion {
                action
                color
                validateOnEnter
              }
            }
            clientMutationId
          }
        }
    """

    inp = {
        "input": extract_serializer_input_fields(
            serializers.SaveActionButtonQuestionSerializer, question
        )
    }
    result = schema_executor(query, variable_values=inp)
    assert not bool(result.errors)
    snapshot.assert_match(result.data)



# -*- coding: utf-8 -*-
# Generated by Django 1.11.20 on 2019-05-10 08:43
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [("caluma_form", "0015_question_format_validators")]

    operations = [
        migrations.AlterField(
            model_name="questionoption",
            name="id",
            field=models.CharField(
                max_length=255, primary_key=True, serialize=False, unique=True
            ),
        )
    ]



import pytest
from django.core.exceptions import ImproperlyConfigured

from caluma.caluma_core.types import Node

from .. import models, serializers
from ..mutation import Mutation
from ..permissions import BasePermission, object_permission_for, permission_for
from .fake_model import get_fake_model


def test_custom_permission_override_has_permission_for_custom_node(
    db, info, history_mock
):
    FakeModel = get_fake_model(model_base=models.UUIDModel)
    instance = FakeModel.objects.create()

    class Serializer(serializers.ModelSerializer):
        class Meta:
            model = FakeModel
            fields = "__all__"

    class CustomMutation(Mutation):
        class Meta:
            serializer_class = Serializer

    class CustomPermission(BasePermission):
        @permission_for(CustomMutation)
        def has_permission_for_custom_mutation(self, mutation, info):
            return False

        @object_permission_for(CustomMutation)
        def has_object_permission_for_custom_mutation(self, mutation, info, instance):
            return False

    assert not CustomPermission().has_permission(CustomMutation, info)
    assert not CustomPermission().has_object_permission(CustomMutation, info, instance)


def test_custom_permission_override_has_permission_with_duplicates(db, info):
    FakeModel = get_fake_model(model_base=models.UUIDModel)

    class Serializer(serializers.ModelSerializer):
        class Meta:
            model = FakeModel
            fields = "__all__"

    class CustomMutation(Mutation):
        class Meta:
            serializer_class = Serializer

    class CustomPermission(BasePermission):
        @permission_for(CustomMutation)
        def has_permission_for_custom_mutation(
            self, mutation, info
        ):  # pragma: no cover
            return False

        @permission_for(CustomMutation)
        def has_permission_for_custom_mutation_2(
            self, mutation, info
        ):  # pragma: no cover
            return False

    with pytest.raises(ImproperlyConfigured):
        CustomPermission()


def test_custom_permission_override_has_object_permission_with_duplicates(db, info):
    FakeModel = get_fake_model(model_base=models.UUIDModel)

    class Serializer(serializers.ModelSerializer):
        class Meta:
            model = FakeModel
            fields = "__all__"

    class CustomMutation(Mutation):
        class Meta:
            serializer_class = Serializer

    class CustomPermission(BasePermission):
        @object_permission_for(CustomMutation)
        def has_object_permission_for_custom_mutation(
            self, mutation, info, instance
        ):  # pragma: no cover
            return False

        @object_permission_for(CustomMutation)
        def has_object_permission_for_custom_mutation_2(
            self, mutation, info, instance
        ):  # pragma: no cover
            return False

    with pytest.raises(ImproperlyConfigured):
        CustomPermission()


def test_custom_permission_override_has_permission_with_multiple_mutations(
    db, history_mock, info
):
    FakeModel = get_fake_model(model_base=models.UUIDModel)

    class Serializer(serializers.ModelSerializer):
        class Meta:
            model = FakeModel
            fields = "__all__"

    class CustomMutation(Mutation):
        class Meta:
            serializer_class = Serializer

    class AnotherMutation(Mutation):
        class Meta:
            serializer_class = Serializer

    class CustomPermission(BasePermission):
        @permission_for(CustomMutation)
        @permission_for(AnotherMutation)
        def has_permission_for_both_mutations(self, mutation, info):  # pragma: no cover
            return False

    assert not CustomPermission().has_permission(CustomMutation, info)
    assert not CustomPermission().has_permission(AnotherMutation, info)


def test_custom_permission_override_has_object_permission_with_multiple_mutations(
    db, history_mock, info
):
    FakeModel = get_fake_model(model_base=models.UUIDModel)
    instance = FakeModel.objects.create()

    class Serializer(serializers.ModelSerializer):
        class Meta:
            model = FakeModel
            fields = "__all__"

    class CustomMutation(Mutation):
        class Meta:
            serializer_class = Serializer

    class AnotherMutation(Mutation):
        class Meta:
            serializer_class = Serializer

    class CustomPermission(BasePermission):
        @object_permission_for(CustomMutation)
        @object_permission_for(AnotherMutation)
        def has_object_permission_for_both_mutations(
            self, mutation, info, instance
        ):  # pragma: no cover
            return False

    assert not CustomPermission().has_object_permission(CustomMutation, info, instance)
    assert not CustomPermission().has_object_permission(AnotherMutation, info, instance)


def test_validate_permission():

    with pytest.raises(ImproperlyConfigured) as exc:

        @object_permission_for(Node)
        def has_object_permission_for_both_mutations(
            self, mutation, info, instance
        ):  # pragma: no cover
            return False

    assert exc.match(
        r"Mutation caluma.caluma_core.types.Node is not a recognized mutation"
    )



import json
from datetime import datetime

import minio.error
import pytest
from django.utils.dateparse import parse_date
from graphql_relay import to_global_id

from caluma.caluma_core.tests import extract_serializer_input_fields
from caluma.caluma_core.validations import BaseValidation

from .. import api, models, serializers
from ..models import Answer, Question


@pytest.mark.parametrize(
    "question__type,answer__value", [(models.Question.TYPE_INTEGER, 23)]
)
def test_remove_answer(db, snapshot, question, answer, schema_executor):
    query = """
        mutation RemoveAnswer($input: RemoveAnswerInput!) {
          removeAnswer(input: $input) {
            answer {
              id
              meta
              __typename
            }
            clientMutationId
          }
        }
    """

    result = schema_executor(
        query, variable_values={"input": {"answer": str(answer.pk)}}
    )

    assert not result.errors
    with pytest.raises(models.Answer.DoesNotExist):
        answer.refresh_from_db()

    snapshot.assert_match(result.data)


@pytest.mark.parametrize(
    "question__type,answer__value", [(models.Question.TYPE_INTEGER, 23)]
)
def test_remove_default_answer(db, snapshot, question, answer, schema_executor):
    query = """
        mutation RemoveDefaultAnswer($input: RemoveDefaultAnswerInput!) {
          removeDefaultAnswer(input: $input) {
            question {
              id
              meta
              __typename
            }
            clientMutationId
          }
        }
    """

    question.default_answer = answer
    question.save()

    result = schema_executor(
        query, variable_values={"input": {"question": str(question.pk)}}
    )

    assert not result.errors
    with pytest.raises(models.Answer.DoesNotExist):
        answer.refresh_from_db()

    snapshot.assert_match(result.data)


@pytest.mark.parametrize("delete_answer", [True, False])
@pytest.mark.parametrize("option__slug", ["option-slug"])
@pytest.mark.parametrize(
    "question__type,answer__value,answer__date,mutation,success",
    [
        (Question.TYPE_INTEGER, 1, None, "SaveDefaultIntegerAnswer", True),
        (Question.TYPE_FLOAT, 2.1, None, "SaveDefaultFloatAnswer", True),
        (Question.TYPE_TEXT, "Test", None, "SaveDefaultStringAnswer", True),
        (Question.TYPE_TEXTAREA, "Test", None, "SaveDefaultStringAnswer", True),
        (Question.TYPE_CHOICE, "option-slug", None, "SaveDefaultStringAnswer", True),
        (
            Question.TYPE_MULTIPLE_CHOICE,
            ["option-slug"],
            None,
            "SaveDefaultListAnswer",
            True,
        ),
        (Question.TYPE_DYNAMIC_CHOICE, "5.5", None, "SaveDefaultStringAnswer", False),
        (
            Question.TYPE_DYNAMIC_MULTIPLE_CHOICE,
            [],
            None,
            "SaveDefaultListAnswer",
            False,
        ),
        (Question.TYPE_DATE, None, "2019-02-22", "SaveDefaultDateAnswer", True),
        (Question.TYPE_FILE, None, None, "SaveDefaultFileAnswer", False),
        (Question.TYPE_TABLE, None, None, "SaveDefaultTableAnswer", True),
    ],
)
def test_save_default_answer_graphql(
    db,
    snapshot,
    question,
    answer,
    mutation,
    question_option,
    document_factory,
    answer_factory,
    answer_document_factory,
    question_factory,
    success,
    schema_executor,
    delete_answer,
    admin_user,
):
    mutation_func = mutation[0].lower() + mutation[1:]
    query = f"""
        mutation {mutation}($input: {mutation}Input!) {{
          {mutation_func}(input: $input) {{
            answer {{
              __typename
              ... on StringAnswer {{
                stringValue: value
              }}
              ... on IntegerAnswer {{
                integerValue: value
              }}
              ... on ListAnswer {{
                listValue: value
              }}
              ... on FloatAnswer {{
                floatValue: value
              }}
              ... on ListAnswer {{
                listValue: value
              }}
              ... on DateAnswer {{
                dateValue: value
              }}
              ... on TableAnswer {{
                table_value: value {{
                  form {{
                    slug
                  }}
                }}
              }}
            }}
            clientMutationId
          }}
        }}
    """

    inp = {
        "input": extract_serializer_input_fields(
            serializers.SaveDefaultAnswerSerializer, answer
        )
    }

    if question.type == Question.TYPE_TABLE:
        documents = document_factory.create_batch(2, form=question.row_form)
        # create a subtree
        sub_question = question_factory(type=Question.TYPE_TEXT)
        document_answer = answer_factory(question=sub_question)
        documents[0].answers.add(document_answer)
        answer_document_factory(answer=answer, document=documents[0])

        inp["input"]["value"] = [str(document.pk) for document in documents]

    if question.type == Question.TYPE_DATE:
        inp["input"]["value"] = answer.date
        answer.value = None
        answer.save()

    question.default_answer = answer
    question.save()

    if delete_answer:
        # delete answer to force create test instead of update
        Answer.objects.filter(pk=answer.pk).delete()

    result = schema_executor(query, variable_values=inp)

    assert not bool(result.errors) == success

    if success:
        snapshot.assert_match(result.data)


@pytest.mark.parametrize("delete_answer", [True, False])
@pytest.mark.parametrize("option__slug", ["option-slug"])
@pytest.mark.parametrize(
    "question__type,answer__value,answer__date,success",
    [
        (Question.TYPE_INTEGER, 1, None, True),
        (Question.TYPE_FLOAT, 2.1, None, True),
        (Question.TYPE_TEXT, "Test", None, True),
        (Question.TYPE_TEXTAREA, "Test", None, True),
        (Question.TYPE_CHOICE, "option-slug", None, True),
        (Question.TYPE_MULTIPLE_CHOICE, ["option-slug"], None, True),
        (Question.TYPE_DYNAMIC_CHOICE, "5.5", None, False),
        (Question.TYPE_DYNAMIC_MULTIPLE_CHOICE, [], None, False),
        (Question.TYPE_DATE, None, "2019-02-22", True),
        (Question.TYPE_FILE, None, None, False),
        (Question.TYPE_TABLE, None, None, True),
    ],
)
def test_save_default_answer_python_api(
    db,
    snapshot,
    question,
    answer,
    question_option,
    document_factory,
    answer_factory,
    answer_document_factory,
    question_factory,
    success,
    delete_answer,
    admin_user,
):
    inp = extract_serializer_input_fields(
        serializers.SaveDefaultAnswerSerializer, answer
    )

    if question.type == Question.TYPE_TABLE:
        documents = document_factory.create_batch(2, form=question.row_form)
        # create a subtree
        sub_question = question_factory(type=Question.TYPE_TEXT)
        document_answer = answer_factory(question=sub_question)
        documents[0].answers.add(document_answer)
        answer_document_factory(answer=answer, document=documents[0])

        inp["value"] = [str(document.pk) for document in documents]

    if question.type == Question.TYPE_DATE:
        inp["value"] = answer.date
        answer.value = None
        answer.save()

        if success:
            inp["value"] = parse_date(inp["value"])

    question.default_answer = answer
    question.save()

    if delete_answer:
        # delete answer to force create test instead of update
        Answer.objects.filter(pk=answer.pk).delete()

    if success:
        answer = api.save_default_answer(question, user=admin_user, value=inp["value"])
        snapshot.assert_match(answer)
    else:
        with pytest.raises(Exception):
            api.save_default_answer(question, user=admin_user, value=inp["value"])


@pytest.mark.parametrize("question__type,answer__value", [(Question.TYPE_FLOAT, 0.1)])
def test_save_calculated_dependency_default_answer(
    db, snapshot, question, answer, question_factory, admin_user
):
    question_factory(
        type=Question.TYPE_CALCULATED_FLOAT,
        calc_expression=f"'{question.slug}'|answer * 10",
    )

    inp = extract_serializer_input_fields(
        serializers.SaveDefaultAnswerSerializer, answer
    )

    question.default_answer = answer
    question.save()

    answer = api.save_default_answer(question, user=admin_user, value=inp["value"])
    snapshot.assert_match(answer)


@pytest.mark.parametrize(
    "question__type,answer__value", [(models.Question.TYPE_TEXT, "foo")]
)
def test_delete_question_with_default(db, question, answer):
    question.default_answer = answer
    question.save()
    question.delete()

    with pytest.raises(models.Answer.DoesNotExist):
        answer.refresh_from_db()

    assert models.Answer.history.count() == 2
    assert all(
        h.history_question_type == question.type for h in models.Answer.history.all()
    )


@pytest.mark.parametrize("question__type", [Question.TYPE_TEXT])
def test_validation_class_save_document_answer(db, mocker, answer, schema_executor):
    class CustomValidation(BaseValidation):
        def validate(self, mutation, data, info):
            data["value"] += " (validated)"
            return data

    mocker.patch(
        "caluma.caluma_form.serializers.SaveAnswerSerializer.validation_classes",
        [CustomValidation],
    )

    query = """
        mutation saveDocumentStringAnswer($input: SaveDocumentStringAnswerInput!) {
          saveDocumentStringAnswer(input: $input) {
            answer {
              __typename
              ... on StringAnswer {
                stringValue: value
              }
            }
            clientMutationId
          }
        }
    """

    variables = {
        "input": {
            "document": to_global_id("StringAnswer", answer.document.pk),
            "question": to_global_id("StringAnswer", answer.question.pk),
            "value": "Test",
        }
    }

    result = schema_executor(query, variable_values=variables)

    assert (
        result.data["saveDocumentStringAnswer"]["answer"]["stringValue"]
        == "Test (validated)"
    )


@pytest.mark.parametrize("question__type", ["file"])
def test_file_answer_metadata(db, answer, schema_executor, minio_mock):
    query = """
        query ans($id: ID!) {
            node(id:$id) {
                ... on FileAnswer {
                    fileValue: value {
                        name
                        downloadUrl
                        metadata
                    }
                }
            }
        }
    """
    vars = {"id": to_global_id("FileAnswer", str(answer.pk))}

    # before "upload"
    old_stat = minio_mock.stat_object.return_value
    minio_mock.stat_object.side_effect = minio.error.S3Error(
        "NoSuchKey",
        "object does not exist",
        resource="bla",
        request_id=134,
        host_id="minio",
        response="bla",
    )

    # Before upload, no metadata is available
    result = schema_executor(query, variable_values=vars)
    assert not result.errors
    assert result.data["node"]["fileValue"]["metadata"] is None

    # After "upload", metadata should contain some useful data
    minio_mock.stat_object.return_value = old_stat
    minio_mock.stat_object.side_effect = None

    result = schema_executor(query, variable_values=vars)
    assert not result.errors

    # Make sure all the values (especially in metadata) are JSON serializable.
    # This is something the schema_executor doesn't test, but may bite us in prod
    metadata = result.data["node"]["fileValue"]["metadata"]
    assert json.dumps(metadata)

    # Ensure some of the important properties having the right types,
    # and being correctly parseable
    assert isinstance(metadata["last_modified"], str)
    assert isinstance(metadata["size"], int)
    assert all(isinstance(m, str) for m in metadata["metadata"].values())
    assert datetime.fromisoformat(metadata["last_modified"])



import json

from graphql.error.syntax_error import GraphQLSyntaxError
from graphql.language import parser, visitor

from caluma.caluma_logging.models import AccessLog


class AccessLogMiddleware:
    def __init__(self, get_response):
        self.get_response = get_response

    def __call__(self, request):
        response = self.get_response(request)
        # user is available only after the request was processed
        try:
            body = json.loads(request.body.decode("utf-8"))
        except json.decoder.JSONDecodeError:  # pragma: no cover
            return response

        vis = AccessLogVisitor()
        try:
            doc = parser.parse(body["query"])
            visitor.visit(doc, vis)
        except GraphQLSyntaxError:
            pass

        AccessLog.objects.create(
            username=request.user.username,
            query=body.get("query"),
            variables=body.get("variables"),
            status_code=response.status_code,
            has_error=response.status_code >= 400,
            **vis.values,
        )

        return response


class AccessLogVisitor(visitor.Visitor):
    def __init__(self):
        self.values = {}

        super().__init__()

    def enter_OperationDefinition(self, node, *args):
        # either "query" or "mutation"
        self.values["operation"] = node.operation
        try:
            self.values["operation_name"] = node.name.value
        except AttributeError:
            pass

    def enter_SelectionSet(self, node, *args):
        # grab name of the query, eg. "allCases", which is the first "selection"
        # thus skip any further selections
        if self.values.get("selection"):
            return
        self.values["selection"] = node.selections[0].name.value



from django.core import management
from django.urls import reverse
from watchman import settings as watchman_settings


def test_db_connection_working(
    disable_logs, track_errors, capsys, client, snapshot, minio_mock_working, db
):
    """Test /healthz/ endpoint response with working database connection.

    Provide database access through the 'db' fixture with working connection.
    Passes health checks for databases (connection), database models and
    database migrations.
    """
    # get /healthz/ response and compare to previous snapshot
    response = client.get(reverse("healthz"))
    snapshot.assert_match(response.json())

    # assert database checks pass
    json_response = response.json()
    assert json_response["databases"] == [{"default": {"ok": True}}]
    assert json_response["database models"] == [{"default": {"ok": True}}]
    assert json_response["database migrations"] == [{"default": {"ok": True}}]
    assert response.status_code == 200

    # assert exception type
    *_, err = capsys.readouterr()
    assert not err


def test_db_connection_broken(
    disable_logs,
    track_errors,
    capsys,
    client,
    snapshot,
    db_broken_connection,
    minio_mock_working,
):
    """Test /healthz/ endpoint response with broken database connection.

    Uses DB fixture with broken connection.
    Affects the health checks for databases (connection), database models and
    database migrations, which should trigger an connection error.
    """
    # get /healthz/ response and compare to previous snapshot
    response = client.get(reverse("healthz"))
    snapshot.assert_match(response.json())

    # assert database checks fail
    json_response = response.json()
    assert json_response["databases"] == [{"default": {"ok": False}}]
    assert json_response["database models"] == [{"default": {"ok": False}}]
    assert json_response["database migrations"] == [{"default": {"ok": False}}]
    assert response.status_code == watchman_settings.WATCHMAN_ERROR_CODE

    # assert exception type
    *_, err = capsys.readouterr()
    assert (
        'django.db.utils.OperationalError: could not translate host name "not-db" to address'
        in err
    )


def test_db_model_accessible(
    disable_logs,
    track_errors,
    capsys,
    client,
    snapshot,
    minio_mock_working,
    db_privileges_working,
):
    """Test /healthz/ endpoint response when necessary DB tables are accessible.

    Provide a database connected through a new test user with the necessary
    privileges to access the tables needed in the health checks.
    Passes database models health check and should be able to access the
    necessary model table.
    """
    # get /healthz/ response and compare to previous snapshot
    response = client.get(reverse("healthz"))
    snapshot.assert_match(response.json())

    # assert database models check passes
    assert response.json()["database models"] == [{"default": {"ok": True}}]
    assert response.status_code == 200

    # assert exception type
    *_, err = capsys.readouterr()
    assert not err


def test_db_model_read_error(
    disable_logs,
    track_errors,
    capsys,
    client,
    snapshot,
    minio_mock_working,
    db_read_error,
):
    """Test /healthz/ endpoint response when necessary DB tables are inaccessible.

    Provide a database connected through a new test user without the necessary
    privileges to access the tables needed in the health checks.
    Affects the database models health check when trying to access the
    specified model table and triggers a read error due to lack of privileges.
    """
    # get /healthz/ response and compare to previous snapshot
    response = client.get(reverse("healthz"))
    snapshot.assert_match(response.json())

    # assert database models check fails
    assert response.json()["database models"] == [{"default": {"ok": False}}]
    assert response.status_code == watchman_settings.WATCHMAN_ERROR_CODE

    # assert exception type
    *_, err = capsys.readouterr()
    assert (
        "django.db.utils.ProgrammingError: permission denied for relation caluma_logging_accesslog"
        in err
    )


def test_db_model_write_error(
    disable_logs,
    track_errors,
    capsys,
    client,
    snapshot,
    minio_mock_working,
    db_write_error,
):
    """Test /healthz/ endpoint response when necessary DB tables are read-only.

    Provide a database connected through a new test user without the necessary
    privileges to write to the tables needed in the health checks.
    Affects the database models health check when trying to create, update or
    delete an object for the specified model table and triggers write error
    due to lack of privileges.
    """
    # get /healthz/ response and compare to previous snapshot
    response = client.get(reverse("healthz"))
    snapshot.assert_match(response.json())

    # assert database models check fails
    assert response.json()["database models"] == [{"default": {"ok": False}}]
    assert response.status_code == watchman_settings.WATCHMAN_ERROR_CODE

    # assert exception type
    *_, err = capsys.readouterr()
    assert (
        "django.db.utils.ProgrammingError: permission denied for relation caluma_logging_accesslog"
        in err
    )


def test_minio_connection_working(
    disable_logs, track_errors, capsys, client, snapshot, db, minio_mock_working
):
    """Test /healthz/ endpoint response when minio connection is working.

    Provide minio mock that passes the media storage service health check.
    """
    # get /healthz/ response and compare to previous snapshot
    response = client.get(reverse("healthz"))
    snapshot.assert_match(response.json())

    # assert media storage service check passes
    assert response.json()["media storage service"] == {"ok": True}
    assert response.status_code == 200

    # assert exception type
    *_, err = capsys.readouterr()
    assert not err


def test_minio_connection_broken(
    disable_logs, track_errors, capsys, client, snapshot, db, minio_broken_connection
):
    """Test /healthz/ endpoint response when minio connection is broken.

    Provide new minio client with faulty connection information.
    Affects media storage service health check and should trigger
    connection error.
    """
    # get /healthz/ response and compare to previous snapshot
    response = client.get(reverse("healthz"))
    snapshot.assert_match(response.json())

    # assert media storage service check fails
    assert response.json()["media storage service"] == {"ok": False}
    assert response.status_code == watchman_settings.WATCHMAN_ERROR_CODE

    # assert exception type
    *_, err = capsys.readouterr()
    assert "urllib3.exceptions.MaxRetryError" in err


def test_minio_not_configured(
    disable_logs, track_errors, capsys, client, snapshot, db, minio_not_configured
):
    """Test /healthz/ endpoint response when minio is not configured.

    Change storage client settings to ensure check assumes no media storage
    service is configured.
    Media storage service health check should pass with a 'not configured'
    annotation.
    """
    # get /healthz/ response and compare to previous snapshot
    response = client.get(reverse("healthz"))
    snapshot.assert_match(response.json())

    # assert media storage service check passes
    assert response.json()["media storage service (not configured)"] == {"ok": True}
    assert response.status_code == 200

    # assert exception type
    *_, err = capsys.readouterr()
    assert not err


def test_minio_read_error(
    disable_logs, track_errors, capsys, client, snapshot, db, minio_mock_read_error
):
    """Test /healthz/ endpoint response when minio receives a read error.

    Mocks minio read functions to produce an error.
    Affects media storage service health check file download and stat_object
    functions.
    """
    # get /healthz/ response and compare to previous snapshot
    response = client.get(reverse("healthz"))
    snapshot.assert_match(response.json())

    # assert media storage service check fails
    assert response.json()["media storage service"] == {"ok": False}
    assert response.status_code == watchman_settings.WATCHMAN_ERROR_CODE

    # assert exception type
    *_, err = capsys.readouterr()
    assert "Exception: stat_object failed" in err


def test_minio_write_error(
    disable_logs, track_errors, capsys, client, snapshot, db, minio_mock_write_error
):
    """Test /healthz/ endpoint response when minio receives a write error.

    Mocks minio write functions to produce an error.
    Affects media storage service health check file upload function.
    """
    # get /healthz/ response and compare to previous snapshot
    response = client.get(reverse("healthz"))
    snapshot.assert_match(response.json())

    # assert media storage service check fails
    assert response.json()["media storage service"] == {"ok": False}
    assert response.status_code == watchman_settings.WATCHMAN_ERROR_CODE

    # assert exception type
    *_, err = capsys.readouterr()
    assert "requests.exceptions.RequestException: PUT request failed" in err


def test_db_migrations_applied(
    disable_logs, track_errors, capsys, client, snapshot, db, minio_mock_working
):
    """Test /healthz/ endpoint response when migrations all created & applied.

    Ensure no changes or unapplied migrations are detected.
    Database migrations health check should pass.
    """
    # Manually migrate database to ensure the database is in the newest
    # state and no unapplied migrations are still around.
    # When running migration tests, database might be left in an
    # inconsistent state.
    management.call_command("migrate")

    # get /healthz/ response and compare to previous snapshot
    response = client.get(reverse("healthz"))
    snapshot.assert_match(response.json())

    # assert database migrations check passes
    assert response.json()["database migrations"] == [{"default": {"ok": True}}]
    assert response.status_code == 200

    # assert exception type
    *_, err = capsys.readouterr()
    assert not err


def test_db_migrations_not_applied(
    disable_logs,
    track_errors,
    capsys,
    client,
    snapshot,
    db,
    minio_mock_working,
    roll_back_migrations,
):
    """Test /healthz/ endpoint response when migrations haven't been applied.

    Undo applied migrations for app 'contenttypes'.
    Affects database migrations health check, which should detect unapplied
    migrations and fail.
    """
    # get /healthz/ response and compare to previous snapshot
    response = client.get(reverse("healthz"))
    snapshot.assert_match(response.json())

    # assert database migrations check fails
    assert response.json()["database migrations"] == [{"default": {"ok": False}}]
    assert response.status_code == watchman_settings.WATCHMAN_ERROR_CODE

    # assert exception type
    *_, err = capsys.readouterr()
    assert "Exception: Database has unapplied migrations (migrate)." in err


def test_healthz_disabled(capsys, client, settings):
    """Assert that /healthz endpoint cannot be retrieved if disabled."""
    # disable healthz endpoint
    settings.ENABLE_HEALTHZ_ENDPOINT = False

    # assert accessing health endpoint results in not found error
    response = client.get(reverse("healthz"))
    assert response.status_code == 404

    # assert exception type
    *_, err = capsys.readouterr()
    assert not err



# -*- coding: utf-8 -*-
# Generated by Django 1.11.20 on 2019-04-04 06:52
from __future__ import unicode_literals

import uuid

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [("caluma_form", "0009_auto_20190321_1722")]

    operations = [
        migrations.CreateModel(
            name="File",
            fields=[
                ("created_at", models.DateTimeField(auto_now_add=True)),
                ("modified_at", models.DateTimeField(auto_now=True)),
                (
                    "created_by_user",
                    models.CharField(blank=True, max_length=150, null=True),
                ),
                (
                    "created_by_group",
                    models.CharField(
                        blank=True, db_index=True, max_length=150, null=True
                    ),
                ),
                (
                    "id",
                    models.UUIDField(
                        default=uuid.uuid4,
                        editable=False,
                        primary_key=True,
                        serialize=False,
                    ),
                ),
                ("name", models.CharField(max_length=255)),
            ],
            options={"abstract": False},
        ),
        migrations.AlterField(
            model_name="question",
            name="type",
            field=models.CharField(
                choices=[
                    ("multiple_choice", "multiple_choice"),
                    ("integer", "integer"),
                    ("float", "float"),
                    ("date", "date"),
                    ("choice", "choice"),
                    ("textarea", "textarea"),
                    ("text", "text"),
                    ("table", "table"),
                    ("form", "form"),
                    ("file", "file"),
                ],
                max_length=15,
            ),
        ),
        migrations.AddField(
            model_name="answer",
            name="file",
            field=models.OneToOneField(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                to="caluma_form.File",
            ),
        ),
    ]



import os

import pytest
from django.conf import settings
from django.core.management import call_command
from minio import Minio
from simple_history.models import registered_models

from caluma.caluma_form.models import Document, Form

from ..management.commands import cleanup_history


def test_create_bucket_command(mocker):
    mocker.patch.object(Minio, "bucket_exists", return_value=False)
    mocker.patch.object(Minio, "make_bucket")
    call_command("create_bucket", stdout=open(os.devnull, "w"))
    Minio.make_bucket.assert_called_once_with(settings.MINIO_STORAGE_MEDIA_BUCKET_NAME)


@pytest.mark.parametrize("force", [True, False])
@pytest.mark.parametrize("dangling", [True, False])
def test_cleanup_history_command(db, force, dangling):
    # we need to override the registered models dict in order to get rid of the
    # fake models created in caluma_core tests
    cleanup_history.registered_models = {
        k: v for k, v in registered_models.items() if not k.startswith("caluma_core_")
    }
    kwargs = {}
    if force:
        kwargs["force"] = force
    if dangling:
        kwargs["dangling"] = dangling

    f1 = Form.objects.create(slug="form 1")

    doc = Document.objects.create(form=f1)
    assert Document.history.count() == 1

    doc.delete()
    f1.delete()

    # created 4 history entries:
    # HistoricalForm: 2 (1 created, 1 deleted)
    # HistoricalDocument: 2 (1 created, 1 deleted)

    call_command("cleanup_history", **kwargs, stdout=open(os.devnull, "w"))

    assert Form.history.count() == 2
    assert Document.history.count() == 0 if (dangling and force) else 2



# -*- coding: utf-8 -*-
# Generated by Django 1.11.20 on 2019-02-20 13:03
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [("caluma_workflow", "0010_auto_20190213_1050")]

    operations = [
        migrations.AddField(
            model_name="task",
            name="is_multiple_instance",
            field=models.BooleanField(
                default=False,
                help_text="Allows creating multiple work items for this task using the `CreateWorkItem` mutation. If true, one work item will be created for each entry in `address_groups`.",
            ),
        ),
        migrations.AlterField(
            model_name="task",
            name="type",
            field=models.CharField(
                choices=[
                    ("simple", "Task which can simply be marked as completed."),
                    (
                        "complete_workflow_form",
                        "Task to complete a defined workflow form.",
                    ),
                    ("complete_task_form", "Task to complete a defined task form."),
                ],
                max_length=50,
            ),
        ),
    ]



# Generated by Django 2.2.12 on 2020-06-19 14:21

from django.db import migrations, models

INITIAL_UPDATE = """
    UPDATE caluma_form_historicalanswer histans
    SET history_question_type = hq_latest.type

    FROM (
        SELECT ha.history_id, hq.type, hq.slug, ha.history_date ha_date, hq.history_date hq_date
        FROM caluma_form_historicalanswer ha
        JOIN caluma_form_historicalquestion hq ON hq.slug = ha.question_id AND hq.history_date = (
            SELECT MAX(hq.history_date) FROM caluma_form_historicalquestion hq WHERE (
                hq.slug = ha.question_id AND hq.history_date <= ha.history_date
            )
        )
    ) hq_latest
    WHERE histans.history_id = hq_latest.history_id;
"""

FALLBACK_UPDATE = """
    UPDATE caluma_form_historicalanswer ha
    SET history_question_type = q.type
    FROM caluma_form_question q
    WHERE ha.history_question_type IS NULL AND q.slug = ha.question_id;
"""


class Migration(migrations.Migration):

    dependencies = [("caluma_form", "0034_fix_fk_lengths")]

    operations = [
        migrations.AddField(
            model_name="historicalanswer",
            name="history_question_type",
            field=models.CharField(default=None, max_length=23, null=True),
            preserve_default=False,
        ),
        migrations.RunSQL(INITIAL_UPDATE, reverse_sql=migrations.RunSQL.noop),
        migrations.RunSQL(FALLBACK_UPDATE, reverse_sql=migrations.RunSQL.noop),
        migrations.AlterField(
            model_name="historicalanswer",
            name="history_question_type",
            field=models.CharField(default="", max_length=23),
            preserve_default=False,
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.11.21 on 2019-06-21 08:55
from __future__ import unicode_literals

import uuid

import django.contrib.postgres.fields
import django.contrib.postgres.fields.jsonb
import django.db.models.deletion
import localized_fields.fields.field
import simple_history.models
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        (
            "caluma_form",
            "0020_historicalanswer_historicalanswerdocument_historicaldocument_historicalfile_historicalform_historica",
        ),
        ("caluma_workflow", "0011_auto_20190220_1303"),
    ]

    operations = [
        migrations.CreateModel(
            name="HistoricalCase",
            fields=[
                ("created_at", models.DateTimeField(blank=True, editable=False)),
                ("modified_at", models.DateTimeField(blank=True, editable=False)),
                (
                    "created_by_user",
                    models.CharField(blank=True, max_length=150, null=True),
                ),
                (
                    "created_by_group",
                    models.CharField(
                        blank=True, db_index=True, max_length=150, null=True
                    ),
                ),
                ("history_user_id", models.CharField(max_length=150, null=True)),
                (
                    "id",
                    models.UUIDField(db_index=True, default=uuid.uuid4, editable=False),
                ),
                (
                    "closed_at",
                    models.DateTimeField(
                        blank=True,
                        help_text="Time when case has either been canceled or completed",
                        null=True,
                    ),
                ),
                (
                    "closed_by_user",
                    models.CharField(blank=True, max_length=150, null=True),
                ),
                (
                    "closed_by_group",
                    models.CharField(blank=True, max_length=150, null=True),
                ),
                (
                    "status",
                    models.CharField(
                        choices=[
                            (
                                "running",
                                "Case is running and work items need to be completed.",
                            ),
                            ("completed", "Case is done."),
                            ("canceled", "Case is cancelled."),
                        ],
                        db_index=True,
                        max_length=50,
                    ),
                ),
                ("meta", django.contrib.postgres.fields.jsonb.JSONField(default=dict)),
                (
                    "history_id",
                    models.UUIDField(
                        default=uuid.uuid4,
                        editable=False,
                        primary_key=True,
                        serialize=False,
                    ),
                ),
                ("history_date", models.DateTimeField()),
                ("history_change_reason", models.CharField(max_length=100, null=True)),
                (
                    "history_type",
                    models.CharField(
                        choices=[("+", "Created"), ("~", "Changed"), ("-", "Deleted")],
                        max_length=1,
                    ),
                ),
                (
                    "document",
                    models.ForeignKey(
                        blank=True,
                        db_constraint=False,
                        null=True,
                        on_delete=django.db.models.deletion.DO_NOTHING,
                        related_name="+",
                        to="caluma_form.Document",
                    ),
                ),
                (
                    "workflow",
                    models.ForeignKey(
                        blank=True,
                        db_constraint=False,
                        null=True,
                        on_delete=django.db.models.deletion.DO_NOTHING,
                        related_name="+",
                        to="caluma_workflow.Workflow",
                    ),
                ),
            ],
            options={
                "verbose_name": "historical case",
                "ordering": ("-history_date", "-history_id"),
                "get_latest_by": "history_date",
            },
            bases=(simple_history.models.HistoricalChanges, models.Model),
        ),
        migrations.CreateModel(
            name="HistoricalFlow",
            fields=[
                ("created_at", models.DateTimeField(blank=True, editable=False)),
                ("modified_at", models.DateTimeField(blank=True, editable=False)),
                (
                    "created_by_user",
                    models.CharField(blank=True, max_length=150, null=True),
                ),
                (
                    "created_by_group",
                    models.CharField(
                        blank=True, db_index=True, max_length=150, null=True
                    ),
                ),
                ("history_user_id", models.CharField(max_length=150, null=True)),
                (
                    "id",
                    models.UUIDField(db_index=True, default=uuid.uuid4, editable=False),
                ),
                ("next", models.TextField()),
                (
                    "history_id",
                    models.UUIDField(
                        default=uuid.uuid4,
                        editable=False,
                        primary_key=True,
                        serialize=False,
                    ),
                ),
                ("history_date", models.DateTimeField()),
                ("history_change_reason", models.CharField(max_length=100, null=True)),
                (
                    "history_type",
                    models.CharField(
                        choices=[("+", "Created"), ("~", "Changed"), ("-", "Deleted")],
                        max_length=1,
                    ),
                ),
            ],
            options={
                "verbose_name": "historical flow",
                "ordering": ("-history_date", "-history_id"),
                "get_latest_by": "history_date",
            },
            bases=(simple_history.models.HistoricalChanges, models.Model),
        ),
        migrations.CreateModel(
            name="HistoricalTask",
            fields=[
                ("created_at", models.DateTimeField(blank=True, editable=False)),
                ("modified_at", models.DateTimeField(blank=True, editable=False)),
                (
                    "created_by_user",
                    models.CharField(blank=True, max_length=150, null=True),
                ),
                (
                    "created_by_group",
                    models.CharField(
                        blank=True, db_index=True, max_length=150, null=True
                    ),
                ),
                ("history_user_id", models.CharField(max_length=150, null=True)),
                ("slug", models.SlugField()),
                ("name", localized_fields.fields.field.LocalizedField(required=[])),
                (
                    "description",
                    localized_fields.fields.field.LocalizedField(
                        blank=True, null=True, required=[]
                    ),
                ),
                (
                    "type",
                    models.CharField(
                        choices=[
                            ("simple", "Task which can simply be marked as completed."),
                            (
                                "complete_workflow_form",
                                "Task to complete a defined workflow form.",
                            ),
                            (
                                "complete_task_form",
                                "Task to complete a defined task form.",
                            ),
                        ],
                        max_length=50,
                    ),
                ),
                ("meta", django.contrib.postgres.fields.jsonb.JSONField(default=dict)),
                (
                    "address_groups",
                    models.TextField(
                        blank=True,
                        help_text="Group jexl returning what group(s) derived work items will be addressed to.",
                        null=True,
                    ),
                ),
                ("is_archived", models.BooleanField(default=False)),
                (
                    "lead_time",
                    models.PositiveIntegerField(
                        blank=True,
                        help_text="Time in seconds task may take to be processed.",
                        null=True,
                    ),
                ),
                (
                    "is_multiple_instance",
                    models.BooleanField(
                        default=False,
                        help_text="Allows creating multiple work items for this task using the `CreateWorkItem` mutation. If true, one work item will be created for each entry in `address_groups`.",
                    ),
                ),
                (
                    "history_id",
                    models.UUIDField(
                        default=uuid.uuid4,
                        editable=False,
                        primary_key=True,
                        serialize=False,
                    ),
                ),
                ("history_date", models.DateTimeField()),
                ("history_change_reason", models.CharField(max_length=100, null=True)),
                (
                    "history_type",
                    models.CharField(
                        choices=[("+", "Created"), ("~", "Changed"), ("-", "Deleted")],
                        max_length=1,
                    ),
                ),
                (
                    "form",
                    models.ForeignKey(
                        blank=True,
                        db_constraint=False,
                        null=True,
                        on_delete=django.db.models.deletion.DO_NOTHING,
                        related_name="+",
                        to="caluma_form.Form",
                    ),
                ),
            ],
            options={
                "verbose_name": "historical task",
                "ordering": ("-history_date", "-history_id"),
                "get_latest_by": "history_date",
            },
            bases=(simple_history.models.HistoricalChanges, models.Model),
        ),
        migrations.CreateModel(
            name="HistoricalTaskFlow",
            fields=[
                ("created_at", models.DateTimeField(blank=True, editable=False)),
                ("modified_at", models.DateTimeField(blank=True, editable=False)),
                (
                    "created_by_user",
                    models.CharField(blank=True, max_length=150, null=True),
                ),
                (
                    "created_by_group",
                    models.CharField(
                        blank=True, db_index=True, max_length=150, null=True
                    ),
                ),
                ("history_user_id", models.CharField(max_length=150, null=True)),
                (
                    "id",
                    models.UUIDField(db_index=True, default=uuid.uuid4, editable=False),
                ),
                (
                    "history_id",
                    models.UUIDField(
                        default=uuid.uuid4,
                        editable=False,
                        primary_key=True,
                        serialize=False,
                    ),
                ),
                ("history_date", models.DateTimeField()),
                ("history_change_reason", models.CharField(max_length=100, null=True)),
                (
                    "history_type",
                    models.CharField(
                        choices=[("+", "Created"), ("~", "Changed"), ("-", "Deleted")],
                        max_length=1,
                    ),
                ),
                (
                    "flow",
                    models.ForeignKey(
                        blank=True,
                        db_constraint=False,
                        null=True,
                        on_delete=django.db.models.deletion.DO_NOTHING,
                        related_name="+",
                        to="caluma_workflow.Flow",
                    ),
                ),
                (
                    "task",
                    models.ForeignKey(
                        blank=True,
                        db_constraint=False,
                        null=True,
                        on_delete=django.db.models.deletion.DO_NOTHING,
                        related_name="+",
                        to="caluma_workflow.Task",
                    ),
                ),
                (
                    "workflow",
                    models.ForeignKey(
                        blank=True,
                        db_constraint=False,
                        null=True,
                        on_delete=django.db.models.deletion.DO_NOTHING,
                        related_name="+",
                        to="caluma_workflow.Workflow",
                    ),
                ),
            ],
            options={
                "verbose_name": "historical task flow",
                "ordering": ("-history_date", "-history_id"),
                "get_latest_by": "history_date",
            },
            bases=(simple_history.models.HistoricalChanges, models.Model),
        ),
        migrations.CreateModel(
            name="HistoricalWorkflow",
            fields=[
                ("created_at", models.DateTimeField(blank=True, editable=False)),
                ("modified_at", models.DateTimeField(blank=True, editable=False)),
                (
                    "created_by_user",
                    models.CharField(blank=True, max_length=150, null=True),
                ),
                (
                    "created_by_group",
                    models.CharField(
                        blank=True, db_index=True, max_length=150, null=True
                    ),
                ),
                ("history_user_id", models.CharField(max_length=150, null=True)),
                ("slug", models.SlugField()),
                ("name", localized_fields.fields.field.LocalizedField(required=[])),
                (
                    "description",
                    localized_fields.fields.field.LocalizedField(
                        blank=True, null=True, required=[]
                    ),
                ),
                ("meta", django.contrib.postgres.fields.jsonb.JSONField(default=dict)),
                ("is_published", models.BooleanField(default=False)),
                ("is_archived", models.BooleanField(default=False)),
                (
                    "allow_all_forms",
                    models.BooleanField(
                        default=False,
                        help_text="Allow workflow to be started with any form",
                    ),
                ),
                (
                    "history_id",
                    models.UUIDField(
                        default=uuid.uuid4,
                        editable=False,
                        primary_key=True,
                        serialize=False,
                    ),
                ),
                ("history_date", models.DateTimeField()),
                ("history_change_reason", models.CharField(max_length=100, null=True)),
                (
                    "history_type",
                    models.CharField(
                        choices=[("+", "Created"), ("~", "Changed"), ("-", "Deleted")],
                        max_length=1,
                    ),
                ),
            ],
            options={
                "verbose_name": "historical workflow",
                "ordering": ("-history_date", "-history_id"),
                "get_latest_by": "history_date",
            },
            bases=(simple_history.models.HistoricalChanges, models.Model),
        ),
        migrations.CreateModel(
            name="HistoricalWorkItem",
            fields=[
                ("created_at", models.DateTimeField(blank=True, editable=False)),
                ("modified_at", models.DateTimeField(blank=True, editable=False)),
                (
                    "created_by_user",
                    models.CharField(blank=True, max_length=150, null=True),
                ),
                (
                    "created_by_group",
                    models.CharField(
                        blank=True, db_index=True, max_length=150, null=True
                    ),
                ),
                ("history_user_id", models.CharField(max_length=150, null=True)),
                (
                    "id",
                    models.UUIDField(db_index=True, default=uuid.uuid4, editable=False),
                ),
                (
                    "closed_at",
                    models.DateTimeField(
                        blank=True,
                        help_text="Time when work item has either been canceled or completed",
                        null=True,
                    ),
                ),
                (
                    "closed_by_user",
                    models.CharField(blank=True, max_length=150, null=True),
                ),
                (
                    "closed_by_group",
                    models.CharField(blank=True, max_length=150, null=True),
                ),
                ("deadline", models.DateTimeField(blank=True, null=True)),
                (
                    "status",
                    models.CharField(
                        choices=[
                            ("ready", "Task is ready to be processed."),
                            ("completed", "Task is done."),
                            ("canceled", "Task is cancelled."),
                        ],
                        db_index=True,
                        max_length=50,
                    ),
                ),
                ("meta", django.contrib.postgres.fields.jsonb.JSONField(default=dict)),
                (
                    "addressed_groups",
                    django.contrib.postgres.fields.ArrayField(
                        base_field=models.CharField(max_length=150),
                        default=list,
                        help_text="Offer work item to be processed by a group of users, such are not committed to process it though.",
                        size=None,
                    ),
                ),
                (
                    "assigned_users",
                    django.contrib.postgres.fields.ArrayField(
                        base_field=models.CharField(max_length=150),
                        default=list,
                        help_text="Users responsible to undertake given work item.",
                        size=None,
                    ),
                ),
                (
                    "history_id",
                    models.UUIDField(
                        default=uuid.uuid4,
                        editable=False,
                        primary_key=True,
                        serialize=False,
                    ),
                ),
                ("history_date", models.DateTimeField()),
                ("history_change_reason", models.CharField(max_length=100, null=True)),
                (
                    "history_type",
                    models.CharField(
                        choices=[("+", "Created"), ("~", "Changed"), ("-", "Deleted")],
                        max_length=1,
                    ),
                ),
                (
                    "case",
                    models.ForeignKey(
                        blank=True,
                        db_constraint=False,
                        null=True,
                        on_delete=django.db.models.deletion.DO_NOTHING,
                        related_name="+",
                        to="caluma_workflow.Case",
                    ),
                ),
                (
                    "child_case",
                    models.ForeignKey(
                        blank=True,
                        db_constraint=False,
                        help_text="Defines case of a sub-workflow",
                        null=True,
                        on_delete=django.db.models.deletion.DO_NOTHING,
                        related_name="+",
                        to="caluma_workflow.Case",
                    ),
                ),
                (
                    "document",
                    models.ForeignKey(
                        blank=True,
                        db_constraint=False,
                        null=True,
                        on_delete=django.db.models.deletion.DO_NOTHING,
                        related_name="+",
                        to="caluma_form.Document",
                    ),
                ),
                (
                    "task",
                    models.ForeignKey(
                        blank=True,
                        db_constraint=False,
                        null=True,
                        on_delete=django.db.models.deletion.DO_NOTHING,
                        related_name="+",
                        to="caluma_workflow.Task",
                    ),
                ),
            ],
            options={
                "verbose_name": "historical work item",
                "ordering": ("-history_date", "-history_id"),
                "get_latest_by": "history_date",
            },
            bases=(simple_history.models.HistoricalChanges, models.Model),
        ),
    ]



import pytest
from django.core.exceptions import ImproperlyConfigured
from django.db.models import CharField, QuerySet

from .. import models
from ..types import DjangoObjectType, Node
from ..visibilities import BaseVisibility, Union, filter_queryset_for
from .fake_model import get_fake_model


def test_custom_visibility_override_filter_queryset_for_custom_node(db, history_mock):
    FakeModel = get_fake_model(model_base=models.UUIDModel)
    FakeModel.objects.create()

    class CustomNode(DjangoObjectType):
        class Meta:
            model = FakeModel

    class CustomVisibility(BaseVisibility):
        @filter_queryset_for(CustomNode)
        def filter_queryset_for_custom_node(self, node, queryset, info):
            return queryset.none()

    queryset = FakeModel.objects
    assert queryset.count() == 1
    queryset = CustomVisibility().filter_queryset(CustomNode, queryset, None)
    assert queryset.count() == 0


def test_custom_visibility_override_filter_queryset_with_duplicates(db, history_mock):
    FakeModel = get_fake_model(model_base=models.UUIDModel)
    FakeModel.objects.create()

    class CustomNode(DjangoObjectType):
        class Meta:
            model = FakeModel

    class CustomVisibility(BaseVisibility):
        @filter_queryset_for(CustomNode)
        def filter_queryset_for_custom_node(
            self, node, queryset, info
        ):  # pragma: no cover
            return queryset.none()

        @filter_queryset_for(CustomNode)
        def filter_queryset_for_custom_node_2(
            self, node, queryset, info
        ):  # pragma: no cover
            return queryset.none()

    with pytest.raises(ImproperlyConfigured):
        CustomVisibility()


def test_custom_node_filter_queryset_improperly_configured(db, history_mock):
    FakeModel = get_fake_model(model_base=models.UUIDModel)
    FakeModel.objects.create()

    class CustomNode(DjangoObjectType):
        visibility_classes = None

        class Meta:
            model = FakeModel

    with pytest.raises(ImproperlyConfigured):
        CustomNode.get_queryset(None, None)


def test_custom_visibility_override_specificity(db, history_mock):
    """The first matching filter 'wins'."""
    FakeModel = get_fake_model(
        dict(name=CharField(max_length=255)), model_base=models.UUIDModel
    )
    FakeModel.objects.create(name="Name1")
    FakeModel.objects.create(name="Name2")

    class CustomNode(DjangoObjectType):
        class Meta:
            model = FakeModel

    class CustomVisibility(BaseVisibility):
        @filter_queryset_for(Node)
        def filter_queryset_for_all(self, node, queryset, info):
            return queryset.none()

        @filter_queryset_for(CustomNode)
        def filter_queryset_for_custom_node(self, node, queryset, info):
            return queryset.filter(name="Name1")

    assert FakeModel.objects.count() == 2
    queryset = CustomVisibility().filter_queryset(Node, FakeModel.objects, None)
    assert queryset.count() == 0
    queryset = CustomVisibility().filter_queryset(CustomNode, FakeModel.objects, None)
    assert queryset.count() == 1


def test_union_visibility(db, history_mock):
    FakeModel = get_fake_model(
        dict(name=CharField(max_length=255)), model_base=models.UUIDModel
    )
    FakeModel.objects.create(name="Name1")
    FakeModel.objects.create(name="Name2")
    FakeModel.objects.create(name="Name3")
    FakeModel.objects.create(name="Name4")

    class CustomNode(DjangoObjectType):
        class Meta:
            model = FakeModel

    class Name1Visibility(BaseVisibility):
        @filter_queryset_for(CustomNode)
        def filter_queryset_for_custom_node(self, node, queryset, info):
            return queryset.filter(name="Name1")

    class Name2Visibility(BaseVisibility):
        @filter_queryset_for(CustomNode)
        def filter_queryset_for_custom_node(self, node, queryset, info):
            return queryset.filter(name="Name2")

    class Name3Visibility(BaseVisibility):
        @filter_queryset_for(CustomNode)
        def filter_queryset_for_custom_node(self, node, queryset, info):
            return queryset.filter(name__in=["Name2", "Name3"])

    class ConfiguredUnion(Union):
        visibility_classes = [Name1Visibility, Name2Visibility, Name3Visibility]

    queryset = FakeModel.objects
    result = Name1Visibility().filter_queryset(CustomNode, queryset, None)
    assert result.count() == 1
    result = Name2Visibility().filter_queryset(CustomNode, queryset, None)
    assert result.count() == 1
    result = Name3Visibility().filter_queryset(CustomNode, queryset, None)
    assert result.count() == 2
    queryset = ConfiguredUnion().filter_queryset(CustomNode, queryset, None)
    assert queryset.count() == 3
    assert queryset.get(name="Name2")


def test_union_visibility_none(db, history_mock):
    FakeModel = get_fake_model(model_base=models.UUIDModel)
    FakeModel.objects.create()

    class CustomNode(DjangoObjectType):
        class Meta:
            model = FakeModel

    class CustomVisibility(BaseVisibility):
        @filter_queryset_for(CustomNode)
        def filter_queryset_for_custom_node(self, node, queryset, info):
            return queryset.none()

    class CustomVisibility2(BaseVisibility):
        @filter_queryset_for(CustomNode)
        def filter_queryset_for_custom_node(self, node, queryset, info):
            return queryset.none()

    class ConfiguredUnion(Union):
        visibility_classes = [CustomVisibility2, CustomVisibility]

    queryset = FakeModel.objects
    result = CustomVisibility().filter_queryset(CustomNode, queryset, None)
    assert result.count() == 0
    result = CustomVisibility2().filter_queryset(CustomNode, queryset, None)
    assert result.count() == 0
    queryset = ConfiguredUnion().filter_queryset(CustomNode, queryset, None)
    assert queryset.count() == 0


def test_union_visibility_return_managers(db, history_mock):
    FakeModel = get_fake_model(
        dict(name=CharField(max_length=255)), model_base=models.UUIDModel
    )
    FakeModel.objects.create(name="Name1")
    FakeModel.objects.create(name="Name2")
    FakeModel.objects.create(name="Name3")
    FakeModel.objects.create(name="Name4")

    class CustomNode(DjangoObjectType):
        class Meta:
            model = FakeModel

    class Name1Visibility(BaseVisibility):
        @filter_queryset_for(Node)
        def filter_queryset_for_node(self, node, queryset, info):
            return queryset

    class Name2Visibility(BaseVisibility):
        @filter_queryset_for(CustomNode)
        def filter_queryset_for_custom_node(self, node, queryset, info):
            return queryset

    class ConfiguredUnion(Union):
        visibility_classes = [Name1Visibility, Name2Visibility]

    queryset = FakeModel.objects
    result = Name1Visibility().filter_queryset(CustomNode, queryset, None)
    assert result.count() == 4
    result = Name2Visibility().filter_queryset(CustomNode, queryset, None)
    assert result.count() == 4
    queryset = ConfiguredUnion().filter_queryset(CustomNode, queryset, None)
    assert queryset.count() == 4
    assert isinstance(queryset, QuerySet)


def test_custom_visibility_chained_decorators(db, history_mock):
    """The first matching filter 'wins'."""
    FakeModel1 = get_fake_model(
        dict(name=CharField(max_length=255)), model_base=models.UUIDModel
    )
    FakeModel1.objects.create(name="Name1")
    FakeModel1.objects.create(name="Name2")

    FakeModel2 = get_fake_model(
        dict(name=CharField(max_length=255)), model_base=models.UUIDModel
    )
    FakeModel2.objects.create(name="Name1")
    FakeModel2.objects.create(name="Name2")

    class CustomNode1(DjangoObjectType):
        class Meta:
            model = FakeModel1

    class CustomNode2(DjangoObjectType):
        class Meta:
            model = FakeModel2

    class CustomVisibility(BaseVisibility):
        @filter_queryset_for(Node)
        def filter_queryset_for_all(self, node, queryset, info):
            return queryset.none()

        @filter_queryset_for(CustomNode1)
        @filter_queryset_for(CustomNode2)
        def filter_queryset_for_custom_node(self, node, queryset, info):
            return queryset.filter(name="Name1")

    assert FakeModel1.objects.count() == 2
    assert FakeModel2.objects.count() == 2
    queryset = CustomVisibility().filter_queryset(Node, FakeModel1.objects, None)
    assert queryset.count() == 0
    queryset = CustomVisibility().filter_queryset(CustomNode1, FakeModel1.objects, None)
    assert queryset.count() == 1
    queryset = CustomVisibility().filter_queryset(CustomNode2, FakeModel2.objects, None)
    assert queryset.count() == 1



import sys
from collections import defaultdict
from datetime import date
from logging import getLogger

from django_filters.constants import EMPTY_VALUES
from rest_framework import exceptions

from caluma.caluma_data_source.data_source_handlers import get_data_sources

from . import jexl, models, structure
from .format_validators import get_format_validators
from .models import DynamicOption, Question

log = getLogger()


class CustomValidationError(exceptions.ValidationError):
    """Custom validation error to carry more information.

    This can carry more information about the error, so the documentValidity
    query can show more useful information.
    """

    def __init__(self, detail, code=None, slugs=None):
        slugs = slugs if slugs else []
        super().__init__(detail, code)
        self.slugs = slugs


class AnswerValidator:
    def _validate_question_text(self, question, value, **kwargs):
        min_length = question.min_length if question.min_length is not None else 0
        max_length = (
            question.max_length if question.max_length is not None else sys.maxsize
        )

        if (
            not isinstance(value, str)
            or len(value) < min_length
            or len(value) > max_length
        ):
            raise CustomValidationError(
                f"Invalid value {value}. "
                f"Should be of type str, and it's length between {min_length} and {max_length}",
                slugs=[question.slug],
            )

    def _validate_question_textarea(self, question, value, **kwargs):
        self._validate_question_text(question, value)

    def _validate_question_float(self, question, value, **kwargs):
        min_value = (
            question.min_value if question.min_value is not None else float("-inf")
        )
        max_value = (
            question.max_value if question.max_value is not None else float("inf")
        )

        if not isinstance(value, float) or value < min_value or value > max_value:
            raise CustomValidationError(
                f"Invalid value {value}. "
                f"Should be of type float, not lower than {min_value} "
                f"and not greater than {max_value}",
                slugs=[question.slug],
            )

    def _validate_question_integer(self, question, value, **kwargs):
        min_value = (
            question.min_value if question.min_value is not None else float("-inf")
        )
        max_value = (
            question.max_value if question.max_value is not None else float("inf")
        )

        if not isinstance(value, int) or value < min_value or value > max_value:
            raise CustomValidationError(
                f"Invalid value {value}. "
                f"Should be of type int, not lower than {min_value} "
                f"and not greater than {max_value}",
                slugs=[question.slug],
            )

    def _validate_question_date(self, question, value, **kwargs):
        if not isinstance(value, date):
            raise CustomValidationError(
                f"Invalid value {value}. Should be of type date.", slugs=[question.slug]
            )

    def _validate_question_choice(self, question, value, **kwargs):
        options = question.options.values_list("slug", flat=True)
        if not isinstance(value, str) or value not in options:
            raise CustomValidationError(
                f"Invalid value {value}. "
                f"Should be of type str and one of the options {', '.join(options)}",
                slugs=[question.slug],
            )

    def _validate_question_multiple_choice(self, question, value, **kwargs):
        options = question.options.values_list("slug", flat=True)
        invalid_options = set(value) - set(options)
        if not isinstance(value, list) or invalid_options:
            raise CustomValidationError(
                f"Invalid options [{', '.join(invalid_options)}]. "
                f"Should be one of the options [{', '.join(options)}]",
                slugs=[question.slug],
            )

    def _validate_question_dynamic_choice(
        self, question, value, document, user, **kwargs
    ):
        if not isinstance(value, str):
            raise CustomValidationError(
                f'Invalid value "{value}". Must be of type str.', slugs=[question.slug]
            )
        self._validate_dynamic_option(question, document, value, user)

    def _validate_dynamic_option(self, question, document, option, user):
        data_source = get_data_sources(dic=True)[question.data_source]
        data_source_object = data_source()

        valid_label = data_source_object.validate_answer_value(
            option, document, question, user
        )
        if valid_label is False:
            raise CustomValidationError(
                f'Invalid value "{option}". Not a valid option.', slugs=[question.slug]
            )

        DynamicOption.objects.get_or_create(
            document=document,
            question=question,
            slug=option,
            defaults={
                "label": valid_label,
                "created_by_user": user.username,
                "created_by_group": user.group,
            },
        )

    def _validate_question_dynamic_multiple_choice(
        self, question, value, document, user, **kwargs
    ):
        if not isinstance(value, list):
            raise CustomValidationError(
                f'Invalid value: "{value}". Must be of type list', slugs=[question.slug]
            )

        for v in value:
            if not isinstance(v, str):
                raise CustomValidationError(
                    f'Invalid value: "{v}". Must be of type string',
                    slugs=[question.slug],
                )
            self._validate_dynamic_option(question, document, v, user)

    def _validate_question_table(
        self, question, value, document, user, instance=None, origin=False, **kwargs
    ):
        if not origin:
            for row_doc in value:
                DocumentValidator().validate(row_doc, user=user, **kwargs)
            return

        # this answer was the entry-point of the validation
        # -> validate row form
        value = value or instance and instance.documents.all() or []
        question = question or instance and instance.question

        for row_doc in value:
            if row_doc.form_id != question.row_form_id:
                raise exceptions.ValidationError(
                    f"Document {row_doc.pk} is not of form type {question.row_form.pk}."
                )

    def _validate_question_file(self, question, value, **kwargs):
        pass

    def _validate_question_calculated_float(self, question, value, **kwargs):
        pass

    def validate(
        self,
        *,
        question,
        document,
        user,
        validation_context=None,
        instance=None,
        origin=False,
        **kwargs,
    ):
        # Check all possible fields for value
        value = None
        for i in ["documents", "file", "date", "value"]:
            value = kwargs.get(i, value)
            if value:
                break

        # empty values are allowed
        # required check will be done in DocumentValidator
        if value:
            validate_func = getattr(self, f"_validate_question_{question.type}")
            validate_func(
                question,
                value,
                document=document,
                user=user,
                validation_context=validation_context,
                instance=instance,
                origin=origin,
            )

        format_validators = get_format_validators(dic=True)
        for validator_slug in question.format_validators:
            format_validators[validator_slug]().validate(value, document)


class DocumentValidator:
    def validate(self, document, user, validation_context=None, **kwargs):
        if not validation_context:
            validation_context = self._validation_context(document)

        self._validate_required(validation_context)

        for answer in document.answers.filter(
            question_id__in=validation_context["visible_questions"]
        ):
            validator = AnswerValidator()
            validator.validate(
                document=document,
                question=answer.question,
                value=answer.value,
                documents=answer.documents.all(),
                user=user,
                validation_context=validation_context,
            )

    def _validation_context(self, document):
        # we need to build the context in two steps (for now), as
        # `self.visible_questions()` already needs a context to evaluate
        # `is_hidden` expressions
        intermediate_context = {
            "form": document.form,
            "document": document,
            "visible_questions": None,
            "jexl_cache": defaultdict(dict),
            "structure": structure.FieldSet(document, document.form),
        }

        intermediate_context["visible_questions"] = self.visible_questions(
            document, intermediate_context
        )
        return intermediate_context

    def visible_questions(self, document, validation_context=None):
        """Evaluate the visibility of the questions for the given context.

        This evaluates the `is_hidden` expression for each question to decide
        if the question is visible.
        Return a list of question slugs that are visible.
        """
        if not validation_context:
            validation_context = self._validation_context(document)
        visible_questions = []

        q_jexl = jexl.QuestionJexl(validation_context)
        for field in validation_context["structure"].children():
            question = field.question
            try:
                is_hidden = q_jexl.is_hidden(field)

                if is_hidden:
                    # no need to descend further
                    continue

                visible_questions.append(question.slug)
                if question.type == Question.TYPE_FORM:
                    # answers to questions in subforms are still in
                    # the top level document
                    sub_context = {**validation_context, "structure": field}
                    visible_questions.extend(
                        self.visible_questions(document, sub_context)
                    )

                elif question.type == Question.TYPE_TABLE:
                    row_visibles = set()
                    # make a copy of the validation context, so we
                    # can reuse it for each row
                    row_context = {**validation_context}
                    for row in field.children():
                        sub_context = {**row_context, "structure": row}
                        row_visibles.update(
                            self.visible_questions(document, sub_context)
                        )
                    visible_questions.extend(row_visibles)

            except (jexl.QuestionMissing, exceptions.ValidationError):
                raise
            except Exception as exc:
                log.error(
                    f"Error while evaluating `is_hidden` expression on question {question.slug}: "
                    f"{question.is_hidden}: {str(exc)}"
                )
                raise RuntimeError(
                    f"Error while evaluating `is_hidden` expression on question {question.slug}: "
                    f"{question.is_hidden}. The system log contains more information"
                )
        return visible_questions

    def _validate_required(self, validation_context):  # noqa: C901
        """Validate the 'requiredness' of the given answers.

        Raise exceptions if a required question is not answered.

        Since we're iterating and evaluating `is_hidden` as well for this
        purpose, we help our call site by returning a list of *non-hidden*
        question slugs.
        """
        required_but_empty = []

        q_jexl = jexl.QuestionJexl(validation_context)
        for field in validation_context["structure"].children():
            question = field.question
            try:
                is_hidden = question.slug not in validation_context["visible_questions"]
                if is_hidden:
                    continue

                # The above `is_hidden` is globally cached per question, mainly to optimize DB access.
                # This means a question could be marked "visible" as it would be visible
                # in another row, but would still be hidden in the local row, if this is a
                # table question context.  Thus, in this case we need to re-evaluate it's
                # hiddenness. Luckily, the JEXL evaluator caches those values (locally).
                with q_jexl.use_field_context(field):
                    if q_jexl.is_hidden(field):
                        continue

                is_required = q_jexl.is_required(field)

                if question.type == Question.TYPE_FORM:
                    # form questions's answers are still in the top level document
                    sub_context = {**validation_context, "structure": field}
                    self._validate_required(sub_context)

                elif question.type == Question.TYPE_TABLE:
                    # We need to validate presence in at least one row, but only
                    # if the table question is required.
                    if is_required and not field.children():
                        raise CustomValidationError(
                            f"no rows in {question.slug}", slugs=[question.slug]
                        )
                    for row in field.children():
                        sub_context = {**validation_context, "structure": row}
                        self._validate_required(sub_context)
                else:
                    value = field.value()
                    if value in EMPTY_VALUES and is_required:
                        required_but_empty.append(question.slug)

            except CustomValidationError as exc:
                required_but_empty.extend(exc.slugs)

            except (jexl.QuestionMissing, exceptions.ValidationError):
                raise
            except Exception as exc:
                log.error(
                    f"Error while evaluating `is_required` expression on question {question.slug}: "
                    f"{question.is_required}: {str(exc)}"
                )

                raise RuntimeError(
                    f"Error while evaluating `is_required` expression on question {question.slug}: "
                    f"{question.is_required}. The system log contains more information"
                )

        if required_but_empty:
            raise CustomValidationError(
                f"Questions {','.join(required_but_empty)} are required but not provided.",
                slugs=required_but_empty,
            )


class QuestionValidator:
    @staticmethod
    def _validate_format_validators(data):
        format_validators = data.get("format_validators")
        if format_validators:
            fv = get_format_validators(include=format_validators, dic=True)
            diff_list = list(set(format_validators) - set(fv))
            if diff_list:
                raise exceptions.ValidationError(
                    f"Invalid format validators {diff_list}."
                )

    @staticmethod
    def _validate_data_source(data_source):
        data_sources = get_data_sources(dic=True)
        if data_source not in data_sources:
            raise exceptions.ValidationError(f'Invalid data_source: "{data_source}"')

    @staticmethod
    def _validate_calc_expression(data):
        """Don't allow other calc questions or non-existent questions."""
        expr = data.get("calc_expression")
        if not expr:
            return

        question_jexl = jexl.QuestionJexl()
        deps = set(question_jexl.extract_referenced_questions(expr))

        calc_slugs = set(
            models.Question.objects.filter(
                pk__in=deps, type=models.Question.TYPE_CALCULATED_FLOAT
            ).values_list("slug", flat=True)
        )
        inexistent_slugs = deps - set(
            models.Question.objects.filter(pk__in=deps).values_list("slug", flat=True)
        )
        illegal_deps = ", ".join(calc_slugs.union(inexistent_slugs))

        if illegal_deps:
            raise exceptions.ValidationError(
                f"Calc expression references other calculated or inexistent questions: {illegal_deps}"
            )

    def validate(self, data):
        if data["type"] in [models.Question.TYPE_TEXT, models.Question.TYPE_TEXTAREA]:
            self._validate_format_validators(data)
        elif data["type"] == models.Question.TYPE_CALCULATED_FLOAT:
            self._validate_calc_expression(data)

        if "dataSource" in data:
            self._validate_data_source(data["dataSource"])


def get_document_validity(document, user):
    validator = DocumentValidator()
    is_valid = True
    errors = []

    try:
        validator.validate(document, user)
    except CustomValidationError as exc:
        is_valid = False
        detail = str(exc.detail[0])
        errors = [{"slug": slug, "error_msg": detail} for slug in exc.slugs]

    return {"id": document.id, "is_valid": is_valid, "errors": errors}



from collections import Iterable

import graphene
from django.core.exceptions import ImproperlyConfigured
from django.db.models.query import QuerySet
from graphene.relay import PageInfo
from graphene.relay.connection import ConnectionField
from graphene_django import types
from graphene_django.fields import DjangoConnectionField
from graphene_django.utils import maybe_queryset

from .pagination import connection_from_list, connection_from_list_slice


class Node(object):
    """Base class to define queryset filters for all nodes."""

    # will be set in caluma_core.AppConfig.ready hook, see apps.py
    # to avoid recursive import error
    visibility_classes = None

    @classmethod
    def get_queryset(cls, queryset, info):
        if cls.visibility_classes is None:
            raise ImproperlyConfigured(
                "check that app `caluma.caluma_core` is part of your `INSTALLED_APPS` "
                "or custom node has `visibility_classes` properly assigned."
            )

        for visibility_class in cls.visibility_classes:
            queryset = visibility_class().filter_queryset(cls, queryset, info)

        return queryset.select_related()


class DjangoObjectType(Node, types.DjangoObjectType):
    """Django object type implementing default get_queryset with visibility layer."""

    class Meta:
        abstract = True


class CountableConnectionBase(graphene.Connection):
    """Connection subclass that supports totalCount."""

    class Meta:
        abstract = True

    total_count = graphene.Int()

    def resolve_total_count(self, info, **kwargs):
        try:
            # DjangoConnectionField sets the length already
            return self.length
        except AttributeError:
            if isinstance(self.iterable, QuerySet):  # pragma: no cover
                return self.iterable.count()
            return len(self.iterable)


class DjangoConnectionField(DjangoConnectionField):
    """
    Custom DjangoConnectionField with fix for hasNextPage/hasPreviousPage.

    This can be removed, when (or better if)
    https://github.com/graphql-python/graphql-relay-py/issues/12
    is resolved.
    """

    @classmethod
    def resolve_connection(cls, connection, args, iterable):
        iterable = maybe_queryset(iterable)
        if isinstance(iterable, QuerySet):
            # only query count on database when pagination is needed
            # resolve_connection may be removed again once following issue is fixed:
            # https://github.com/graphql-python/graphene-django/issues/177
            if "before" in args or "after" in args or "first" in args or "last" in args:
                _len = iterable.count()
            else:
                _len = len(iterable)
        else:  # pragma: no cover
            _len = len(iterable)
        connection = connection_from_list_slice(
            iterable,
            args,
            slice_start=0,
            list_length=_len,
            list_slice_length=_len,
            connection_type=connection,
            edge_type=connection.Edge,
            pageinfo_type=PageInfo,
        )
        connection.iterable = iterable
        connection.length = _len
        return connection


class ConnectionField(ConnectionField):
    """
    Custom ConnectionField with fix for hasNextPage/hasPreviousPage.

    This can be removed, when (or better if)
    https://github.com/graphql-python/graphql-relay-py/issues/12
    is resolved.
    """

    @classmethod
    def resolve_connection(cls, connection_type, args, resolved):
        if not resolved:
            resolved = []
        if isinstance(resolved, connection_type):  # pragma: no cover
            return resolved

        assert isinstance(resolved, Iterable), (
            "Resolved value from the connection field have to be iterable or instance of {0}. "
            'Received "{1}"'
        ).format(connection_type, resolved)
        connection = connection_from_list(
            resolved,
            args,
            connection_type=connection_type,
            edge_type=connection_type.Edge,
            pageinfo_type=PageInfo,
        )
        connection.iterable = resolved
        return connection



from django.db import migrations
from caluma.caluma_form.models import Question


def remove_value(apps, schema_editor):
    apps.get_model("caluma_form", "Answer").objects.filter(
        question__type=Question.TYPE_TABLE, value__isnull=False
    ).update(value=None)


class Migration(migrations.Migration):
    dependencies = [("caluma_form", "0037_default_answer_one2one")]
    operations = [migrations.RunPython(remove_value, migrations.RunPython.noop)]



from datetime import timedelta

import pytest
from django.core.management import call_command
from django.utils import timezone

from caluma.caluma_logging.models import AccessLog


@pytest.mark.parametrize("force", [True, False])
@pytest.mark.parametrize(
    "keep,exist",
    [
        (None, 10),
        ("2 weeks", 10),
        ("1 month", 10),
        ("1 day", 1),
        ("7 days", 7),
        ("0 days", 0),
    ],
)
def test_cleanup_access_log(db, access_log_factory, force, keep, exist):
    for i in range(10):
        entry = access_log_factory()
        entry.timestamp = timezone.now() - timedelta(days=i)
        entry.save()

    kwargs = {"force": force}
    if keep:
        kwargs["keep"] = keep

    call_command("cleanup_access_log", **kwargs)

    count = AccessLog.objects.all().count()

    if force:
        assert count == exist
    else:
        assert count == 10
